"""
ATLK with partial observability evaluation functions.

These functions are based on the fully symbolic approach of Xuang and van der
Meyden.
"""

import gc
from functools import reduce
from collections import OrderedDict

from pynusmv.dd import BDD, dynamic_reordering_enabled, reorder
from pynusmv.fsm import BddTrans
from pynusmv.mc import eval_simple_expression
from pynusmv.utils import fixpoint as fp
from pynusmv import node, glob
from pynusmv.exception import PyNuSMVError

from pynusmv_lower_interface.nusmv.compile.symb_table import symb_table as nssymb_table

from ..atlkFO.ast import (TrueExp, FalseExp, Init, Reachable,
                          Atom, Not, And, Or, Implies, Iff, 
                          AF, AG, AX, AU, AW, EF, EG, EX, EU, EW,
                          nK, nE, nD, nC, K, E, D, C,
                          CEF, CEG, CEX, CEU, CEW, CAF, CAG, CAX, CAU, CAW)

from ..atlkFO.eval import nk, ne, nc

from . import config


def evalATLK(fsm, spec, variant="SF", semantics="group"):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    
    fsm -- a MAS representing the system
    spec -- an AST-based ATLK specification
    variant -- the variant of the algorithm to evaluate strategic operators;
               must be
               * "SF" for the standard way: encoding strategies for the
                 agents then performing the fully symbolic model checking
               * "FSF" for the filter-split-filter way: filtering winning
                 sub-protocol before encoding it and performing the fully
                 symbolic model checking
    semantics -- the semantics to use for starting point and strategy point
                 equivalence; must be
                 * "group" for the original ATLK_irF semantics considering
                   the group as a single agent (distributed knowledge is used)
                 * "individual" for the original ATL_ir semantics considering
                   the group as individual agents (individual knowledge is
                   used)
                 
    If variant is not in {"SF", "FSF"}, the standard "SF" way is used.
    If semantics is not in {"group", "individual"}, the "group" semantics is
    used.
    """
    
    if type(spec) is TrueExp:
        return BDD.true(fsm.bddEnc.DDmanager)
        
    elif type(spec) is FalseExp:
        return BDD.false(fsm.bddEnc.DDmanager)
        
    elif type(spec) is Init:
        return fsm.init
        
    elif type(spec) is Reachable:
        return fsm.reachable_states
    
    elif type(spec) is Atom:
        return eval_simple_expression(fsm, spec.value)
        
    elif type(spec) is Not:
        return ~evalATLK(fsm, spec.child, variant=variant)
        
    elif type(spec) is And:
        return (evalATLK(fsm, spec.left, variant=variant)
                & evalATLK(fsm, spec.right, variant=variant))
        
    elif type(spec) is Or:
        return (evalATLK(fsm, spec.left, variant=variant)
                | evalATLK(fsm, spec.right, variant=variant))
        
    elif type(spec) is Implies:
        # a -> b = ~a | b
        return ((~evalATLK(fsm, spec.left, variant=variant))
                 | evalATLK(fsm, spec.right, variant=variant))
        
    elif type(spec) is Iff:
        # a <-> b = (a & b) | (~a & ~b)
        l = evalATLK(fsm, spec.left, variant=variant)
        r = evalATLK(fsm, spec.right, variant=variant)
        return (l & r) | ((~l) & (~r))
        
    elif type(spec) is EX:
        return ex(fsm, evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is AX:
        # AX p = ~EX ~p
        return ~ex(fsm, ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is EG:
        return ew(fsm, evalATLK(fsm, spec.child, variant=variant),
                       BDD.false(fsm.bddEnc.DDmanager))
        
    elif type(spec) is AG:
        # AG p = ~EF ~p = ~E[ true U ~p ]
        return ~eu(fsm,
                   BDD.true(fsm.bddEnc.DDmanager),
                   ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is EU:
        return eu(fsm, evalATLK(fsm, spec.left, variant=variant),
                       evalATLK(fsm, spec.right, variant=variant))
        
    elif type(spec) is AU:
        # A[p U q] = ~E[~q W ~p & ~q] = ~(E[~q U ~p & ~q] | EG ~q)
        p = evalATLK(fsm, spec.left, variant=variant)
        q = evalATLK(fsm, spec.right, variant=variant)
        return ~ew(fsm, ~q, ~p & ~q)
        
    elif type(spec) is EF:
        # EF p = E[ true U p ]
        return eu(fsm,
                  BDD.true(fsm.bddEnc.DDmanager),
                  evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is AF:
        # AF p = ~EG ~p
        return ~ew(fsm, ~evalATLK(fsm, spec.child, variant=variant),
                        BDD.false(fsm.bddEnc.DDmanager))
        
    elif type(spec) is EW:
        # E[ p W q ] = E[ p U q ] | EG p
        return ew(fsm, evalATLK(fsm, spec.left, variant=variant),
                        evalATLK(fsm, spec.right, variant=variant))
        
    elif type(spec) is AW:
        # A[p W q] = ~E[~q U ~p & ~q]
        p = evalATLK(fsm, spec.left, variant=variant)
        q = evalATLK(fsm, spec.right, variant=variant)
        return ~eu(fsm, ~q, ~p & ~q)
        
    elif type(spec) is nK:
        return nk(fsm, spec.agent.value,
                  evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is K:
        # K<'a'> p = ~nK<'a'> ~p
        return ~nk(fsm, spec.agent.value,
                   ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is nE:
        return ne(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is E:
        # E<g> p = ~nE<g> ~p
        return ~ne(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is nD:
        return nd(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child, variant=variant)) 
        
    elif type(spec) is D:
        # D<g> p = ~nD<g> ~p
        return ~nd(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is nC:
        return nc(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is C:
        # C<g> p = ~nC<g> ~p
        return ~nc(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child, variant=variant))
    
    elif type(spec) is CAX:
        # [g] X p = ~<g> X ~p
        newspec = CEX(spec.group, Not(spec.child))
        return ~evalATLK(fsm, newspec, variant=variant)
        
    elif type(spec) is CAG:
        # [g] G p = ~<g> F ~p
        newspec = CEF(spec.group, Not(spec.child))
        return ~evalATLK(fsm, newspec, variant=variant)
        
    elif type(spec) is CAU:
        # [g][p U q] = ~<g>[ ~q W ~p & ~q ]
        newspec = CEW(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right)))
        return ~evalATLK(fsm, newspec, variant=variant)
        
    elif type(spec) is CAF:
        # [g] F p = ~<g> G ~p
        newspec = CEG(spec.group, Not(spec.child))
        return ~evalATLK(fsm, newspec, variant=variant)
        
    elif type(spec) is CAW:
        # [g][p W q] = ~<g>[~q U ~p & ~q]
        newspec = CEU(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right)))
        return ~evalATLK(fsm, newspec, variant=variant)
                   
    elif type(spec) in {CEX, CEG, CEU, CEF, CEW}:
        if variant == "SF":
            return eval_strat(fsm, spec, semantics=semantics)
        elif variant == "FSF":
            return eval_strat_FSF(fsm, spec, semantics=semantics)
        else:
            return eval_strat(fsm, spec, semantics=semantics)
        
    else:
        # TODO Generate error
        print("[ERROR] evalATLK: unrecognized specification type", spec)
        return None


def _fair(fsm):
    if len(fsm.fairness_constraints) <= 0:
        return BDD.true(fsm.bddEnc.DDmanager)
    else:
        run = fsm.trans
        # fair = nu Z. /\_fc pre(mu Y. (Z /\ fc) \/ pre(Y))
        def inner(Z):
            res = BDD.true(fsm.bddEnc.DDmanager)
            for fc in fsm.fairness_constraints:
                res = res & run.pre(fp(lambda Y: (Z & fc) | run.pre(Y),
                                       BDD.false(fsm.bddEnc.DDmanager)))
            return res
        return fp(inner, BDD.true(fsm.bddEnc.DDmanager))


def ex(fsm, phi):
    run = fsm.trans
    return run.pre(phi & _fair(fsm))

def eu(fsm, phi, psi):
    run = fsm.trans
    return fp(lambda Y : (psi & _fair(fsm)) | (phi & run.pre(Y)),
              BDD.false(fsm.bddEnc.DDmanager))


def ew(fsm, phi, psi):
    run = fsm.trans
    if len(fsm.fairness_constraints) <= 0:
        return fp(lambda Z: psi | (phi & run.pre(Z)),
                  BDD.true(fsm.bddEnc.DDmanager))
    else:
        def inner(Z):
            res = BDD.true(fsm.bddEnc.DDmanager)
            for f in fsm.fairness_constraints:
                res = res & run.pre(fp(lambda Y: (psi & _fair(fsm)) |
                                                 (Z & f) | phi & run.pre(Y),
                                    BDD.false(fsm.bddEnc.DDmanager)))
            return (res & phi) | (psi & _fair(fsm))
        return fp(inner, BDD.true(fsm.bddEnc.DDmanager))


def cex_si(fsm, agents, phi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents> X phi
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    strat -- a BDD representing allowed state/inputs pairs, or None
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask
    
    return fsm.pre_strat_si(phi | nfair_gamma_si(fsm, agents, strat),
                            agents, strat)
    

def ceu_si(fsm, agents, phi, psi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying
    <agents>[phi U psi]
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    psi = psi & fsm.bddEnc.statesInputsMask & strat
    
    if len(fsm.fairness_constraints) == 0:
        return fp(lambda Z : psi | (phi & fsm.pre_strat_si(Z, agents, strat)),
                  BDD.false(fsm.bddEnc.DDmanager))
    else:
        nfair = nfair_gamma_si(fsm, agents, strat)
        def inner(Z):
            res = psi
            for f in fsm.fairness_constraints:
                nf = ~f & fsm.bddEnc.statesMask & strat
                res = res | fsm.pre_strat_si(fp(lambda Y :
                                                 (phi | psi | nfair) &
                                                 (Z | nf) &
                                                 (psi |
                                                  fsm.pre_strat_si(Y, agents,
                                                                   strat)
                                                 ),
                                              BDD.true(fsm.bddEnc.DDmanager)),
                                              agents, strat)
            return (psi | phi | nfair) & res
        return fp(inner, BDD.false(fsm.bddEnc.DDmanager))
    

def cew_si(fsm, agents, phi, psi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents>[phi W psi]
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    psi = psi & fsm.bddEnc.statesInputsMask & strat
    
    nfair = nfair_gamma_si(fsm, agents, strat)
    
    return fp(lambda Y : (psi | phi | nfair) &
                         (psi | fsm.pre_strat_si(Y, agents, strat)),
              BDD.true(fsm.bddEnc.DDmanager))
    
    
def ceg_si(fsm, agents, phi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents> G phi
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    
    nfair = nfair_gamma_si(fsm, agents, strat)
    
    return fp(lambda Y : (phi | nfair) & fsm.pre_strat_si(Y, agents, strat),
              BDD.true(fsm.bddEnc.DDmanager))


def nfair_gamma_si(fsm, agents, strat=None):
    """
    Return the set of state/inputs pairs of strat
    in which agents can avoid a fair path in strat.
    If strat is None, it is considered true.
    
    fsm -- the model
    agents -- a list of agents names
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    if len(fsm.fairness_constraints) == 0:
        return BDD.false(fsm.bddEnc.DDmanager)
    else:
        def inner(Z):
            res = BDD.false(fsm.bddEnc.DDmanager)
            for f in fsm.fairness_constraints:
                nf = ~f & fsm.bddEnc.statesMask & strat
                res = res | fsm.pre_strat_si(fp(lambda Y :
                                                 (Z | nf) &
                                                 fsm.pre_strat_si(Y, agents,
                                                                  strat),
                                             BDD.true(fsm.bddEnc.DDmanager)),
                                             agents, strat)
            return res
        return fp(inner, BDD.false(fsm.bddEnc.DDmanager))


def _ifair(fsm, name):
    """
    Return a mu-calculus translation of fair states of fsm under uniform
    strategy played by name.
    """
    
    jump = fsm.transitions[name + "_jump"]
    equiv = fsm.transitions[name + "_equiv"]
    follow = fsm.transitions[name + "_follow"]
    
    # ifair =
    # nu Z. /\_fc <>_group_follow(mu Y. (Z /\ fc) \/ <>_group_follow(Y))
    def inner(Z):
        # /\_fc <>_group_follow(mu Y. (Z /\ fc) \/ <>_group_follow(Y))
        res = BDD.true(fsm.bddEnc.DDmanager)
        for f in fsm.fairness_constraints:
            res = res & follow.pre(fp(lambda Y: (Z & f) | follow.pre(Y),
                                      BDD.false(fsm.bddEnc.DDmanager)))
        return res
    res = fp(inner, BDD.true(fsm.bddEnc.DDmanager))
    
    return res


def _nfair(fsm, name):
    """
    Return a mu-calculus translation of non-fair states of fsm under uniform
    strategy played by name.
    """
    
    jump = fsm.transitions[name + "_jump"]
    equiv = fsm.transitions[name + "_equiv"]
    follow = fsm.transitions[name + "_follow"]
    
    # nfair =
    # mu Z. \/_fc []_group_follow(nu Y. (Z \/ ~fc) /\ []_group_follow(Y))
    def inner(Z):
        # \/_fc []_group_follow(nu Y. (Z \/ ~fc) /\ []_group_follow(Y))
        res = BDD.false(fsm.bddEnc.DDmanager)
        for f in fsm.fairness_constraints:
            res = res | ~follow.pre(~fp(lambda Y: (Z | ~f) & ~follow.pre(~Y),
                                       BDD.true(fsm.bddEnc.DDmanager)))
        return res
    res = fp(inner, BDD.false(fsm.bddEnc.DDmanager))
    return res


def cex_symbolic(fsm, name, child):
    jump = fsm.transitions[name + "_jump"]
    equiv = fsm.transitions[name + "_equiv"]
    follow = fsm.transitions[name + "_follow"]
    reachable = fsm.reachable_states
    
    if len(fsm.fairness_constraints) <= 0:
        # <group>i X p =
        # <>_group_jump []_group_equiv (reachable => []_group_follow p)
        return jump.pre(~equiv.pre(~(reachable.imply(~follow.pre(~child)))))
    else:
        # <group>i X fair p =
        # <>_group_jump []_group_equiv
        # (reachable => []_group_follow (p \/ ~ifair))
        return jump.pre(~equiv.pre(~(reachable.imply(
                        ~follow.pre(~(child | _nfair(fsm, name)))))))


def ceu_symbolic(fsm, name, left, right):
    jump = fsm.transitions[name + "_jump"]
    equiv = fsm.transitions[name + "_equiv"]
    follow = fsm.transitions[name + "_follow"]
    reachable = fsm.reachable_states
    
    if len(fsm.fairness_constraints) <= 0:
        # <group>i[p U q] =
        # <>_group_jump []_group_equiv
        # (reachable => mu Z . q \/ (p /\ []_group_follow Z))
        return jump.pre(~equiv.pre(~(reachable.imply(
                        fp(lambda Z: right | (left & ~follow.pre(~Z)),
                           BDD.false(fsm.bddEnc.DDmanager))))))
    else:
        nfair = _nfair(fsm, name)
        # <group>i fair [p U q] =
        # <>_group_jump []_group_equiv
        # reachable =>
        # mu Z. (p \/ q \/ ~ifair) /\
        #       (q \/_fc []_group_follow
        #                (nu Y. (Z \/ ~fc) /\
        #                              (p \/ q \/ ~ifair) /\
        #                              (q \/ []_group_follow(Y))))
        def inner(Z):
            # (p \/ q \/ ~ifair) /\
            # (q \/_fc []_group_follow
            #          (nu Y. (Z \/ ~fc) /\
            #                        (p \/ q \/ ~ifair) /\
            #                        (q \/ []_group_follow(Y))))
            res = BDD.false(fsm.bddEnc.DDmanager)
            for f in fsm.fairness_constraints:
                # []_group_follow (nu Y. (Z \/ ~fc) /\
                #                        (p \/ q \/ ~ifair) /\
                #                        (q \/ []_group_follow(Y)))
                res = res | ~follow.pre(
                      ~(fp(lambda Y: ((Z | ~f) & (left | right | nfair) &
                                     (right | ~follow.pre(~Y))),
                           BDD.true(fsm.bddEnc.DDmanager))))
            return ((left | right | nfair) & (right | res))
        return jump.pre(~equiv.pre(~(reachable.imply(
                        fp(inner, BDD.false(fsm.bddEnc.DDmanager))))))


def cew_symbolic(fsm, name, left, right):
    jump = fsm.transitions[name + "_jump"]
    equiv = fsm.transitions[name + "_equiv"]
    follow = fsm.transitions[name + "_follow"]
    reachable = fsm.reachable_states
    
    if len(fsm.fairness_constraints) <= 0:
        # <group>i[p W q] = 
        # <>_group_jump []_group_equiv
        # (reachable => nu Z . q \/ (p /\ []_group_follow Z))
        return jump.pre(~equiv.pre(~(reachable.imply(
                        fp(lambda Z: right | (left & ~follow.pre(~Z)),
                           BDD.true(fsm.bddEnc.DDmanager))))))
    else:
        nfair = _nfair(fsm, name)
        # <group>i fair [p W q] =
        # <>_group_jump []_group_equiv
        # reachable =>
        # nu Z. (p \/ q \/ ~ifair) /\ (q \/ []_group_follow(Z))
        res = jump.pre(~equiv.pre(~(reachable.imply(
                        fp(lambda Z: ((left | right | nfair) &
                                      (right | ~follow.pre(~Z))),
                           BDD.true(fsm.bddEnc.DDmanager))))))
        return res


def get_equiv_class(fsm, gamma, state, semantics="group"):
    """
    Return the equivalence class for gamma state belongs to. The equivalence
    class depends on semantics; if semantics is individual, the common
    knowledge based equivalence class for gamma containing state is returned;
    otherwise (assuming semantics is group), the distributed knowledge based
    equivalence fo gamma containing state is returned.
    
    fsm -- the model
    gamma -- a set of agents of fsm
    semantics -- the semantic to use (group or individual);
                 if not "individual", "group" semantics is used
    """
    if semantics == "individual":
        old = BDD.false(fsm.bddEnc.DDmanager)
        eq = state
        while old != eq:
            old = eq
            for agent in gamma:
                eq |= (fsm.equivalent_states(old, {agent}) &
                       fsm.reachable_states)
            
        return eq
    else:
        return fsm.equivalent_states(state, gamma) & fsm.reachable_states


def is_conflicting(fsm, eqclass, gamma, semantics="group"):
    """
    Return whether the given equivalence class for gamma is conflicting or not.
    
    fsm -- the model
    eqclass -- a BDD representing a set of states/inputs pairs
    gamma -- a set of agents of fsm
    semantics -- the semantic to use for splitting (group or individual);
                 if not "individual", "group" semantics is used
    
    If semantics is "individual", eqclass must be a group knowledge based
    equivalence class for gamma; otherwise, semantics is assumed to be "group"
    and eqclass must be a distributed knowledge based equivalence class.
    
    """
    if semantics == "individual":
        # Consider each agent separately
        for agent in gamma:
            nagent_cube = (fsm.bddEnc.inputsCube -
                           fsm.inputs_cube_for_agents({agent}))
            agentclass = eqclass
            # Get one equivalence class for agent at a time
            while agentclass.isnot_false():
                si = fsm.pick_one_state_inputs(agentclass)
                s = si.forsome(fsm.bddEnc.inputsCube)
                eqcl = agentclass & get_equiv_class(fsm, {agent}, s)
                
                # Check if the equivalence class is conflicting for agent
                if ((eqcl - (eqcl & si.forsome(fsm.bddEnc.statesCube |
                                               nagent_cube)))
                     .isnot_false()):
                    return True
                
                agentclass -= eqcl
        
        return False
    else:
        si = fsm.pick_one_state_inputs(eqclass)
        ngamma_cube = fsm.bddEnc.inputsCube - fsm.inputs_cube_for_agents(gamma)
        return (eqclass - (eqclass & si.forsome(fsm.bddEnc.statesCube |
                                                ngamma_cube))).isnot_false()

def split_for_one_agent(fsm, strat, agent):
    """
    Split the given set of moves strat into non-conflicting sub-strategies for
    agent.
    
    fsm -- the model
    strat -- a BDD representing a set of states/inputs pairs
    agent -- an agent of fsm
    
    The strategies are restricted to moves of strat instead of general
    strategies for agent.
    """
    
    if strat.is_false():
        yield strat
    
    else:
        nagent_cube = (fsm.bddEnc.inputsCube -
                       fsm.inputs_cube_for_agents({agent}))
        
        s = fsm.pick_one_state(strat)
        eqclass = get_equiv_class(fsm, {agent}, s) & strat
        # semantics here has no impact since individual meets group semantics
        # with one single agent
        
        strat = strat - eqclass
        
        while eqclass.isnot_false():
            si = fsm.pick_one_state_inputs(eqclass)
            ia = si.forsome(fsm.bddEnc.statesCube | nagent_cube)
            ncss = eqclass & ia
            
            eqclass -= ncss
            
            for sstrat in split_for_one_agent(fsm, strat, agent):
                yield sstrat | ncss


def split_conflicting(fsm, eqclass, gamma, semantics="group"):
    """
    Split the given equivalence class for gamma into non conflicting
    sub-strategies.
    
    fsm -- the model
    eqclass -- a BDD representing a set of states/inputs pairs
    gamma -- a set of agents of fsm
    semantics -- the semantic to use for splitting (group or individual);
                 if not "individual", "group" semantics is used
    
    If semantics is "individual", eqclass must be a group knowledge based
    equivalence class for gamma; otherwise, semantics is assumed to be "group"
    and eqclass must be a distributed knowledge based equivalence class.
    
    Generate all splitted non conflicting sub-strategies.
    
    """
    if semantics == "individual":
        if len(gamma) <= 0:
            yield eqclass
        else:
            agent = next(iter(sorted(gamma)))
            agents = gamma - {agent}
            for sstrat in split_conflicting(fsm, eqclass, agents,
                                            semantics="individual"):
                for strat in split_for_one_agent(fsm, sstrat, agent):
                    yield strat
    else:
        ngamma_cube = fsm.bddEnc.inputsCube - fsm.inputs_cube_for_agents(gamma)
        # Split eqclass into non-conflicting subsets
        while eqclass.isnot_false():
            si = fsm.pick_one_state_inputs(eqclass)
            ncss = eqclass & si.forsome(fsm.bddEnc.statesCube | ngamma_cube)
            eqclass = eqclass - ncss

            yield ncss


def split_one(fsm, strats, gamma, semantics="group"):
    """
    Split one equivalence class of strats and return triples composed of
    the common non-conflicting part already encountered,
    a split and the rest of strats to split.
    
    fsm -- the model
    strats -- a BDD representing a set of states/inputs pairs
    gamma -- a set of agents of fsm
    semantics -- the semantic to use for splitting (group or individual);
                 if not "individual", "group" semantics is used
    
    Return a generator of all triples of common parts, splits
    and rest of strats.
    
    """
    if strats.is_false():
        yield (strats, strats, strats)
        return
        
    else:
        common = BDD.false(fsm.bddEnc.DDmanager)
        
        while strats.isnot_false():
            # Get one equivalence class
            si = fsm.pick_one_state_inputs(strats)
            s = si.forsome(fsm.bddEnc.inputsCube)
            eqs = get_equiv_class(fsm, gamma, s, semantics=semantics)
            eqcl = strats & eqs
            
            # Remove it from strats
            strats = strats - eqcl
            
            # The current equivalence class is conflicting
            if is_conflicting(fsm, eqcl, gamma, semantics=semantics):
                for non_conflicting in split_conflicting(fsm, eqcl, gamma,
                                                         semantics=semantics):
                    yield (common, non_conflicting, strats)
                return
            
            else:
                # Add equivalence class to common
                common = common | eqcl
        
        # strats is false, everything is in common
        yield (common, strats, strats)


def split(fsm, strats, gamma, semantics="group"):
    """
    Split strats into all its non-conflicting greatest subsets.
    
    fsm -- the model
    strats -- a BDD representing a set of states/inputs pairs
    gamma -- a set of agents of fsm
    semantics -- the semantic to use for splitting (group or individual)
    
    Return a generator of all non-conflicting greatest subsets of strats.
    
    """
    if strats.is_false():
        yield strats
    else:
        # semantics should be group
        # Split one equivalence class
        for common, splitted, rest in split_one(fsm, strats, gamma,
                                                semantics=semantics):
            for strat in split(fsm, rest, gamma,
                               semantics=semantics):
                yield (common | strat | splitted)


def filter_strat(fsm, spec, strat=None, variant="SF", semantics="group"):
    """
    Returns the subset SA of strat (or the whole system if strat is None),
    state/action pairs of fsm, such that there is a strategy to satisfy spec
    in fsm.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator;
            the operator is CEX, CEG, CEF, CEU or CEW.
    strat -- the subset of the system to consider.
    variant -- the variant of the algorithm to evaluate strategic operators;
               must be
               * "SF" for the standard way: splitting in uniform strategies then
                 filtering winning states,
               * "FS" for the alternating way: filtering winning states, then
                 splitting one conflicting equivalence class, then recurse
               * "FSF" for the filter-split-filter way: filtering winning states
                 then splitting all remaining actions into uniform strategies,
                 then filtering final winning states.
    semantics -- the semantics to use for starting point and strategy point
                 equivalence; must be
                 * "group" for the original ATLK_irF semantics considering
                   the group as a single agent (distributed knowledge is used)
                 * "individual" for the original ATL_ir semantics considering
                   the group as individual agents (individual knowledge is
                   used)
                 
    If variant is not in {"SF", "FS", "FSF"}, the standard "SF" way is used.
    """
    
    sat = BDD.false(fsm.bddEnc.DDmanager)
    agents = {atom.value for atom in spec.group}
    
    if strat is None:
        strat = BDD.true(fsm.bddEnc.DDmanager)
        # We can be more restrictive here: strat = fsm.protocol(agents)
    
    # Filtering
    if type(spec) is CEX:
        winning = cex_si(fsm, agents,
                         evalATLK(fsm, spec.child, variant=variant,
                                  semantics=semantics), strat)

    elif type(spec) is CEG:
        winning = ceg_si(fsm, agents, evalATLK(fsm, spec.child,
                                               variant=variant,
                                               semantics=semantics), strat)

    elif type(spec) is CEU:
        winning = ceu_si(fsm, agents, evalATLK(fsm, spec.left,
                                               variant=variant,
                                               semantics=semantics),
                         evalATLK(fsm, spec.right, variant=variant,
                                  semantics=semantics), strat)

    elif type(spec) is CEF:
        # <g> F p = <g>[true U p]
        winning = ceu_si(fsm, agents, BDD.true(fsm.bddEnc.DDmanager),
                         evalATLK(fsm, spec.child, variant=variant,
                                  semantics=semantics), strat)

    elif type(spec) is CEW:
       winning = cew_si(fsm, agents, evalATLK(fsm, spec.left, variant=variant,
                                              semantics=semantics),
                        evalATLK(fsm, spec.right, variant=variant,
                                 semantics=semantics), strat)
    
    
    return winning & fsm.bddEnc.statesInputsMask & fsm.protocol(agents)


def all_equiv_sat(fsm, winning, agents, semantics="group"):
    """
    Return the states s of winning such that all states indistinguishable
    from s by agents are in winning.
    
    fsm -- a MAS representing the system;
    winning -- a BDD representing a set of states of fsm;
    agents -- a set of agents;
    semantics -- the semantics to use (group or individual); if not
                 "individual", "group" semantics is used.
    
    """
    if semantics == "individual":
        equiv_sat = BDD.true(fsm.bddEnc.DDmanager)
        for agent in agents:
            equiv_sat &= all_equiv_sat(fsm, winning, {agent},
                                       semantics="group")
        return equiv_sat
    else:
        # Get states for which all states belong to winning
        # wineq is the set of states for which all equiv states are in winning
        nwinning = ~winning & fsm.bddEnc.statesInputsMask
        return ~(fsm.equivalent_states(nwinning &
                 fsm.reachable_states, frozenset(agents))) & winning


def vars_trans_from_protocol(fsm, original_variables, original_trans,
                             strat, observables, actions, name):
    """
    Given the BDD fsm of the original model, the set of original variables
    (as triplets of variable name, variable type, variable kind) and
    the transition relation of the original model,
    return the new variables to declare and new transition relations to define
    to fully encode strat in the model, given that strat define a set of
    strategies based on observables and actions (both strings of NuSMV
    state and inputs variables, respectively).
    
    The returned value is a pair where the first element is an enumerable
    of (variable names (as Nodes), variable type (as SymbType),
    variable kind) triplets and the second element is a dictionary of
    trans-name (as str) -> trans-expression.
    
    name is used to prefix variables representing the strategies.
    """
    
    def _var_to_name(variable):
        """
        Given a variable of the model, return a name that can be used as an
        atom for the variable name.
        """
        return (str(variable).replace(".", "$").replace("[", "$").
                replace("]", "$"))

    def _assign_to_name(assigment):
        """
        Given an assigment of variables (an enumerable of var, val pairs),
        return a string that can be used as an atom for representing the
        assigment.
        """
        return "##".join(_var_to_name(var) + "#" + val
                         for var, val in sorted(((str(var), str(val))
                                                 for var, val in assigment)))
    
    variables = OrderedDict()
    act_cube = fsm.bddEnc.cube_for_inputs_vars(actions)
    obs_cube = fsm.bddEnc.cube_for_state_vars(observables)
    state_cube = fsm.bddEnc.statesCube
    inputs_cube = fsm.bddEnc.inputsCube
    other_vars_cube = (state_cube) - obs_cube
    other_acts_cube = (inputs_cube) - act_cube
    
    # Get all observations by manipulating the protocol
    # and for each observation
    # get the possible values for agent's actions
    protocol = (strat & fsm.bddEnc.statesInputsMask & fsm.state_constraints &
                fsm.inputs_constraints)
    while protocol.isnot_false():
        si = fsm.pick_one_state_inputs(protocol)
        state = fsm.pick_one_state(si.forsome(fsm.bddEnc.inputsCube))
        
        # Get observation values
        obs_vars = tuple(sorted([(node.Identifier.from_string(var),
                                  node.Expression.from_string(val))
                                 for var, val 
                                 in state.get_str_values().items()
                                 if var in observables]))
        
        variables[obs_vars] = list()
        inputs = ((protocol &
                   state.forsome(other_vars_cube)).forsome(
                   fsm.bddEnc.statesCube)
                  & fsm.bddEnc.statesInputsMask
                  & fsm.bddEnc.statesMask
                  & fsm.bddEnc.inputsMask
                  & fsm.state_constraints
                  & fsm.inputs_constraints)
        # Get the possible values for actions
        while inputs.isnot_false():
            i = fsm.pick_one_inputs(inputs)
            
            inputs_vars = tuple(sorted(
                          [(node.Identifier.from_string(var),
                            node.Expression.from_string(val))
                           for var, val in i.get_str_values().items()
                           if var in actions]))
            variables[obs_vars].append(inputs_vars)
            
            inputs = inputs - i.forsome(other_acts_cube)
        
        protocol = (protocol - state.forsome(other_vars_cube))
    
    # Translate observations into state variables
    #   => var#value##var#value...
    # translate values for actions in enums
    #   => ivar#value##ivar#value...
    
    
    new_variables = []
    strategies = list()  # strategies is a list of tuples
                         #   (obs assigment,
                         #    corresponding strategy var,
                         #    corresponding strategy var value,
                         #    actions assigment)
    
    # Compute the number of encoded strategies:
    # the number of encoded strategies is the product of the number of all
    # values of each variable encoding the strategy
    # Prod_{obs} |values(obs)|
    
    nb_encoded_strategies = 1
    
    for obs_values in variables:
        var_name = (name + "_" + _assign_to_name(obs_values))
        var_name = node.Identifier.from_string(var_name)
        type_values = []
        for action_values in variables[obs_values]:
            type_value = _assign_to_name(action_values)
            type_values.append(type_value)
            
            strategies.append((obs_values,
                               var_name,
                               type_value,
                               action_values))
        
        nb_encoded_strategies *= len(type_values)
        
        var_type = fsm.bddEnc.symbTable._get_type_from_node(
                   node.Scalar(type_values))
        
        # a triplet (variable name (as Nodes),
        #            variable type (as SymbType),
        #            variable kind (as int))
        new_variables.append((var_name,
                              var_type,
                              nssymb_table.SYMBOL_STATE_VAR))
    
    if config.debug:
        print("Symbolic: {} strategies encoded for {}.".format(
              nb_encoded_strategies, name))
    
    new_variables = sorted(new_variables, key=lambda e: str(e[0]))
    
    strategies = sorted(strategies, key=lambda e: str(e[1]))
    
    # Relations:
    #   (1) one to follow the strategy (name_follow)
    #       =>  original trans
    #           & use the actions in the strategy
    #           & the strategies stays the same
    #   (2) one to consider equivalent states (name_equiv)
    #       =>  observations stay the same
    #           & the strategies stays the same
    #   (3) one to jump from a strategy to another (name_jump)
    #       => all state variables stay the same
    
    new_trans = OrderedDict()
    
    # name_follow = original_trans & strategy stays the same
    #               & the actions are followed
    
    # original trans
    trans = original_trans
    
    # strategy stays the same: variables are in new_variables
    for var_name, _, _ in new_variables:
        trans &= (var_name.next() == var_name)
    
    # actions are followed:
    #   observations are true
    #   & corresponding strategy var is ivar#value##...
    #   ->
    #   ivar = value & ...
    # strategies is a set of tuples
    #   (obs assigment,
    #    corresponding strategy var,
    #    corresponding strategy var value,
    #    actions assigment)
    for move in strategies:
        obs_assign, strat_var, strat_val, act_assign = move
        
        trans_cond = reduce(lambda e, n: e & n,
                            [var == val for var, val in obs_assign],
                            node.Trueexp())
        trans_cond = trans_cond & (strat_var == strat_val)
        
        trans_cons = reduce(lambda e, n: e & n,
                            [var == val for var, val in act_assign],
                            node.Trueexp())
        trans = trans & trans_cond.implies(trans_cons)
    
    new_trans[name + "_follow"] = trans
    
    
    # name_equiv = observations stay the same
    #              & the strategy stays the same
    trans = reduce(lambda e, n: e & n,
                   [node.Identifier.from_string(str(obs)) ==
                    node.Identifier.from_string(str(obs)).next()
                    for obs in observables], node.Trueexp())
    for var_name, _, _ in new_variables:
        trans &= (var_name.next() == var_name)
    
    new_trans[name + "_equiv"] = trans
    
    
    # name_jump = all state variables stay the same
    trans = []
    for var_name, _, kind in original_variables:
        if kind is nssymb_table.SYMBOL_STATE_VAR:
            trans.append(var_name.next() == var_name)
    trans = reduce(lambda e, n: e & n, trans)
    
    new_trans[name + "_jump"] = trans
    
    return (new_variables, new_trans)

def encode_vars_trans(fsm, name, new_variables, new_trans):
    """
    Encode the new variables in new_variables and the new transition relation
    in new_trans.
    
    The result is
        - the variables in new_variables have been declared and encoded in a
          new layer named name
        - fsm.transitions has been populated with the transition relations of
          new_trans
    """
    if not hasattr(fsm, "transitions"):
        fsm.transitions = {}
    
    symb_table = fsm.bddEnc.symbTable
    new_layer = name
    symb_table.create_layer(new_layer)
    
    # Encode variables
    for var, type_, kind in new_variables:
        if kind is nssymb_table.SYMBOL_STATE_VAR:
            symb_table.declare_state_var(new_layer, var, type_)
        elif kind is nssymb_table.SYMBOL_FROZEN_VAR:
            symb_table.declare_frozen_var(new_layer, var, type_)
        elif kind is nssymb_table.SYMBOL_INPUT_VAR:
            symb_table.declare_input_var(new_layer, var, type_)
    glob.encode_variables_for_layers(layers=[new_layer])
    
    # Create and store transition relations (equiv, jump, follow)
    for trans_name, trans_expr in new_trans.items():
        trans = BddTrans.from_string(symb_table, str(trans_expr))
        fsm.transitions[trans_name] = trans
    
    # Sentinel value for specifying that transitions for group_name have been
    # computed
    fsm.transitions[name] = None

def encode_strat(fsm, agents, name, strat=None, semantics="group"):
    """
    Encode the strategies of the given agents in fsm under the given name;
    if strat is not None, the strategies of agents are restricted to strat
    
    The result is
        - the variables necessary to encode strat have been declared and
          encoded in a new layer named name
        - fsm.transitions has been populated by three new transition relations:
            1. name_follow is the transition relation encoding the fact that
               the current strategy is followed (that is moves are followed
               and the strategy stays the same)
            2. name_equiv is the transition relation encoding the equivalence
               classes of observables, while conserving the current strategy
            3. name_jump is the transtion relation encoding the jump from
               one strategy to another, keeping the current state intact
    
    Note: we assume that the flat hierarchy of fsm is glob.flat_hierarchy()
    """
    
    if strat is None:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    if semantics == "individual":
        if not hasattr(fsm, "transitions"):
            fsm.transitions = {}
        
        # TODO Do not re-encode variables for agents already encoded
        
        flat = glob.flat_hierarchy()
        symb_table = fsm.bddEnc.symbTable
        new_layer = name
        
        # Get original transition relation and original variables
        original_trans = flat.trans
        original_variables = []
        for variable in flat.variables:
            original_variables.append((variable,
                                       symb_table.get_variable_type(variable),
                                       nssymb_table.
                                       SymbTable_get_symbol_category
                                       (symb_table._ptr, variable._ptr)))
        
        new_variables = []
        new_trans = {}
        for agent in agents:
            agent_strat = (fsm.protocol({agent}) &
                           strat.forsome(fsm.bddEnc.inputsCube -
                                         fsm.bddEnc.cube_for_inputs_vars(
                                         fsm.agents_inputvars[agent])))
            nvars, ntrans = vars_trans_from_protocol(fsm,
                                       original_variables, original_trans,
                                       agent_strat,
                                       fsm.agents_observed_variables[agent],
                                       fsm.agents_inputvars[agent], name)
            
            new_variables += list(nvars)
            for trans_name, trans in ntrans.items():
                if trans_name not in new_trans:
                    new_trans[trans_name] = trans
                else:
                    new_trans[trans_name] &= trans

        encode_vars_trans(fsm, name, new_variables, new_trans)
        
        # Force reordering if enabled
        if dynamic_reordering_enabled():
            reorder(method="same")
        
    else:
        # Warning: no determinism of variable order
        agents_actions = set()
        for agent in agents:
            agents_actions |= set(fsm.agents_inputvars[agent])
        agents_obs = set()
        for agent in agents:
            agents_obs |= set(fsm.agents_observed_variables[agent])
        
        agents_strat = fsm.protocol(agents) & strat
        
        if not hasattr(fsm, "transitions"):
            fsm.transitions = {}
        
        flat = glob.flat_hierarchy()
        symb_table = fsm.bddEnc.symbTable
        new_layer = name
        
        # Get original transition relation and original variables
        original_trans = flat.trans
        original_variables = []
        for variable in flat.variables:
            original_variables.append((variable,
                                       symb_table.get_variable_type(variable),
                                       nssymb_table.
                                       SymbTable_get_symbol_category
                                       (symb_table._ptr, variable._ptr)))
        
        # Get variables and transition relations
        new_variables, new_trans = vars_trans_from_protocol(fsm,
                                   original_variables, original_trans,
                                   agents_strat, agents_obs, agents_actions,
                                   name)
        
        encode_vars_trans(fsm, name, new_variables, new_trans)


def agents_in_group(fsm, group):
    """
    Return the list of agents in the given group in fsm. If group is not a
    group of fsm, group is returned alone.
    
    This procedure recursively searches for all basic agents in the groups
    possibly composing group.
    
    fsm -- the model;
    group -- a group or agent name of the model.
    """
    
    if group in fsm.groups:
        agents = []
        for agent in fsm.groups[group]:
            if agent in fsm.groups:
                agents += [a for a in agents_in_group(fsm, agent)
                           if a not in agents]
            else:
                agents.append(agent)
    else:
        agents = [group]
    return agents


def agents_in_list(fsm, agents):
    """
    Return the list of agents in the given list of groups and agents.
    
    fsm -- the model;
    agents -- an iterable of groups and agents.
    """
    result = []
    for agent in agents:
        result += [a for a in agents_in_group(fsm, agent)
                   if a not in result]
    return result


def eval_strat(fsm, spec, semantics="group"):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    spec is a strategic operator <G> pi.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator.
    semantics -- the semantics to use for starting point and strategy point
                 equivalence; must be
                 * "group" for the original ATLK_irF semantics considering
                   the group as a single agent (distributed knowledge is used)
                 * "individual" for the original ATL_ir semantics considering
                   the group as individual agents (individual knowledge is
                   used)
    """
    
    agents = [atom.value for atom in spec.group]
    
    if semantics == "individual":
        agents = agents_in_list(fsm, agents)
    
    group_name = semantics + "_" + "_".join(sorted(agents))
    
    # Create the transition relations for agents if needed
    if not hasattr(fsm, "transitions") or group_name not in fsm.transitions:
        
        if config.debug:
            print("Fully symbolic approach: encoding strategies for {}."
                  .format(group_name))
        
        encode_strat(fsm, agents, group_name, semantics=semantics)

    # Compute the set of winning states
    if config.debug:
        print("Fully symbolic approach: computing winning states for {}."
              .format(spec))
    
    if type(spec) is CEX:
        win = cex_symbolic(fsm, group_name,
                           evalATLK(fsm, spec.child, variant="SF",
                                    semantics=semantics))

    elif type(spec) is CEG:
        win = cew_symbolic(fsm, group_name,
                           evalATLK(fsm, spec.child, variant="SF",
                                    semantics=semantics),
                           BDD.false(fsm.bddEnc.DDmanager))

    elif type(spec) is CEU:
        win = ceu_symbolic(fsm, group_name,
                           evalATLK(fsm, spec.left, variant="SF",
                                    semantics=semantics),
                           evalATLK(fsm, spec.right, variant="SF",
                                    semantics=semantics))

    elif type(spec) is CEF:
        win = ceu_symbolic(fsm, group_name,
                           BDD.true(fsm.bddEnc.DDmanager),
                           evalATLK(fsm, spec.child, variant="SF",
                                    semantics=semantics))

    elif type(spec) is CEW:
        win = cew_symbolic(fsm, group_name,
                           evalATLK(fsm, spec.left, variant="SF",
                                    semantics=semantics),
                           evalATLK(fsm, spec.right, variant="SF",
                                    semantics=semantics))
    
    if config.debug:
        print("Fully symbolic approach: winning states for {} computed."
              .format(spec))
    
    return win


def spec_to_group_name(spec):
    """Return a name that can be used as a NuSMV identifier, based on spec."""
    return "spec" + (str(spec)
                     .replace(" ", "_")
                      .replace(",", "_")
                     .replace("&", "and")
                     .replace("|", "or")
                     .replace("<", "")
                     .replace(">", "")
                     .replace("~", "")
                     .replace("'", "")
                     .replace("=", "")
                     .replace("(", "")
                     .replace(")", "")
                     .replace("[", "")
                     .replace("]", "")
                     .replace("!", "")
                    )
    
def eval_strat_FSF(fsm, spec, semantics="group"):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    spec is a strategic operator <G> pi.
    
    Implement a variant of the algorithm that
    filters, splits and then filters.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator.
    semantics -- the semantics to use for starting point and strategy point
                 equivalence; must be
                 * "group" for the original ATLK_irF semantics considering
                   the group as a single agent (distributed knowledge is used)
                 * "individual" for the original ATL_ir semantics considering
                   the group as individual agents (individual knowledge is
                   used)
    
    """
    
    # For every sub-formula we build a new set of transition relations
    # TODO evaluate whether building a set of transition relations per
    # group of agents is better (this means limiting oneself to a set of
    # sub-formulas...)
    
    group_name = semantics + "_" + spec_to_group_name(spec)
    
    # Create the transition relations for agents if needed
    if not hasattr(fsm, "transitions") or group_name not in fsm.transitions:
        agents = [atom.value for atom in spec.group]

        if semantics == "individual":
            agents = agents_in_list(fsm, agents)

        strat = filter_strat(fsm, spec, strat=fsm.protocol(agents),
                             variant="FSF")

        if strat.is_false():
            return strat
        
        encode_strat(fsm, agents, group_name, strat=strat,
                     semantics=semantics)

    # Compute the set of winning states
    if type(spec) is CEX:
        winning = cex_symbolic(fsm, group_name,
                               evalATLK(fsm, spec.child, variant="FSF"))

    elif type(spec) is CEG:
        winning = cew_symbolic(fsm, group_name,
                               evalATLK(fsm, spec.child, variant="FSF"),
                               BDD.false(fsm.bddEnc.DDmanager))

    elif type(spec) is CEU:
        winning = ceu_symbolic(fsm, group_name,
                               evalATLK(fsm, spec.left, variant="FSF"),
                               evalATLK(fsm, spec.right, variant="FSF"))

    elif type(spec) is CEF:
        winning = ceu_symbolic(fsm, group_name,
                               BDD.true(fsm.bddEnc.DDmanager),
                               evalATLK(fsm, spec.child, variant="FSF"))

    elif type(spec) is CEW:
        winning = cew_symbolic(fsm, group_name,
                               evalATLK(fsm, spec.left, variant="FSF"),
                               evalATLK(fsm, spec.right, variant="FSF"))
    
    winning = (winning.forsome(fsm.bddEnc.inputsCube)
               & fsm.reachable_states)
    return winning


from pynusmv.be.expression import Be 

def model_problem(fsm, bound):
    """
    Computes the unrolled transition relation [[M]]_{k}
    
    .. note:: 
        this is equivalent to :see:`pynusmv.bmc.utils.BmcModel.path(bound)`
    
    :param fsm: the fsm whose transition relation must be unrolled
    :param bound: the bound up to which the transition relation must be unrolled
    :return: the unrolled transition relation of `fsm`
    """
    enc   = fsm.encoding
    # initial state
    init0 = enc.shift_to_time(fsm.init, 0)
    # transition relation (unrolled k steps)
    trans = Be.true(enc.manager)
    for k in range(bound):
        trans = trans & enc.shift_to_time(fsm.trans, k)
    
    return init0 & trans

def invariants_constraint(fsm, k):
    """
    Computes the constraint to be added to the model when one wants to 
    enforce the invariants declared in the SMV text.
    
    .. note:: 
        This constraint has no counterpart in BmcModel. However, the same
        result can be achieved by computing the conjunction of 
        BmcModel.invar[time] for all the times in the range [0; k]. 
    
    :param fsm: the fsm whose transition relation must be unrolled
    :param k: the maximum (horizon/bound) time of the problem
    :return: the invariants constraint enforced on the paths.
    """
    # I know, thisone is REALLY easy!
    # note: the and_interval is just a shortcut for
    #    \bigwedge_{i=0}^{k} (invariants_{i})
    return fsm.encoding.and_interval(fsm.invariants, 0, k)

def generate_problem(fml, fsm, k=10, no_fairness=False, no_invar=False):
    """
    Generates a formula representing a SAT problem that is satisfiable iff
    the the `fsm` violates the formula represented in `formula_text`.
    
    Mathematically, this function computes the following formula:
    .. math:: [[M, f]]_{k}
    
    :param fml: an LTL formula parsed with `pynusmv_tools.bmcLTL.parsing` (hence the 
        abstract syntax tree of that formula). Note, this is *NOT* the NuSMV
        format (Node).
    :param fsm: the FSM representing the model.
    :param k: the maximum (horizon/bound) time of the problem
    :param no_fairness: a flag telling whether or not the generated problem should
        focus on fair executions only (the considered fairness constraints must
        be declared in the SMV model).
    :param no_invar: a flag telling whether or not the generated problem 
        should enforce the declared invariants (these must be declared in the
        SMV text).
    :return: a Be expression that is satisfiable iff the fsm can violate the 
        stated property [[M, f]]_{k}
    """
    negated = fml.nnf(True)
    
    problem = model_problem(fsm, k) & \
              negated.bounded_semantics(fsm, k, fairness = not no_fairness)
    
    # enforce invariants if needed
    if not no_invar:
        problem &= invariants_constraint(fsm, k)
        
    return problem


from pynusmv.dd import BDD

from .eval import ex, eg, eu, fair_states
from pynusmv_tools.explanation.explanation import Explanation


def explain_one_succ(fsm, state, sat):
    """
    Explain why state can reach sat in one step.
    The returned value is a tuple (action, successor)
    where successor is a successor of state through action belonging to sat.
    
    fsm -- the fsm;
    state -- a state of fsm able to reach sat in one step;
    sat -- a BDD representing a set of states of fsm.    
    """
    next = fsm.pick_one_state(fsm.post(state) & sat)
    inputs = fsm.pick_one_inputs(fsm.get_inputs_between_states(state, next))
    return (inputs, next)
    

def explain_all_succ(fsm, state, sat):
    """
    Give all the successors of state belonging to sat.
    The returned value is a set of tuples (action, successor)
    where successor is a successor of state through action belonging to sat.
    
    fsm -- the fsm;
    state -- a state of fsm able to reach sat in one step;
    sat -- a BDD representing a set of states of fsm.    
    """
    successors = set()
    for next in fsm.pick_all_states(fsm.post(state) & sat):
        inputs = fsm.pick_one_inputs(fsm.get_inputs_between_states(state, next))
        successors.add((inputs, next))
    return successors
    

def explain_ex(fsm, state, phi):
    """
    Explain why state of fsm satisfies EX phi.
    
    fsm -- the fsm;
    state -- a state of fsm satisfying EX phi;
    phi -- the set of states of fsm satifying phi.
    """
    # To show that state satisfies EX phi, we have to exhibit a successor
    # of state belonging to phi that is reachable and fair.
    
    # EX phi = fsm.pre(phi & fair)
    
    # We don't explain why the extracted successor is fair.
    
    phi = phi & fair_states(fsm) & fsm.reachable_states
    
    # Get successor of state belonging to phi
    inputs, next = explain_one_succ(fsm, state, phi)
    
    return Explanation(state, {(inputs, Explanation(next))})
    
    
def explain_eu(fsm, state, phi, psi):
    """
    Explain why state of fsm satisfies E phi U psi.
    
    fsm -- the fsm;
    state -- a state of fsm satisfying E phi U psi;
    phi -- the set of states of fsm satifying phi;
    psi -- the set of states of fsm satifying psi.
    """
    # E phi U psi = mu Z . (psi & fair) | (phi & fsm.pre(Z))
    
    # To show that state satisfies E phi U psi
    # we have to show a path of successors satisfying phi,
    # ending in a fair state satisfying psi.
    
    # If state satisfies psi, stop;
    # otherwise, compute the fixpoint until state is in the last set,
    # then show that state can reach the previous set,
    # and recursively explain why the reached state satisfies E phi U psi.
    
    # We don't explain why the psi-states are fair.
    
    if state <= psi:
        return Explanation(state)
        
    else:
        # Compute the fixpoint until state is reached (backward)
        fair = fair_states(fsm)
        f = lambda Y : (psi & fair) | (phi & fsm.pre(Y))
        old = BDD.false(fsm.bddEnc.DDmanager)
        new = psi
        while not state <= new:
            old = new
            new = f(old)
            
        # If state <= new, (and it was not the case in the previous iteration)
        # old does not contain state but a successor of state.
        
        # explain that state can reach old
        inputs, next = explain_one_succ(fsm, state, old)
        successors = {(inputs, explain_eu(fsm, next, phi, psi))}
        
        return Explanation(state, successors)
    
    
def explain_eg(fsm, state, phi):
    """
    Explain why state of fsm satisfies EG phi.
    
    fsm -- the fsm;
    state -- a state of fsm satisfying EG phi;
    phi -- the set of states of fsm satifying phi.
    """
    # To explain why state satisfies EG phi,
    # we have to show a path ending with a loop going through all fairness
    # constraints
    
    # If no fairness constraints are provided,
    # model checking sums up to
    # EG p = nu Z. p & Pre(Z)
    # In this case, just explain by extending EX EG p until a former state is
    # encountered
    if len(fsm.fairness_contraints) <= 0:
        states = BDD.false(fsm.bddEnc.DDmanager)
        orig_expl = expl = Explanation(state)
        eg_phi = eg(fsm, phi)
        while not state <= states:
            states = states | state
            if (fsm.post(state) & states).isnot_false():
                # There is a successor of state that we already saw.
                # Use this one
                action, state = explain_one_succ(fsm, state, states)
            else:
                # There is no successor of state in states, guess a successor
                action, state = explain_one_succ(fsm, state, eg_phi)
            nextexpl = Explanation(state)
            expl.successors = {(action, nextexpl)}
            expl = nextexpl
        # state <= states
        # this means that state lead to a previous state in the path
        # we must explain it
        cur = orig_expl
        while cur.state != state:
            cur = cur.successors[0][1]
        action = fsm.pick_one_inputs(
                                fsm.get_inputs_between_states(state, cur.state))
        expl.successors = {(action, cur)}
        return orig_expl
    
    # Otherwise, there are fairness constraints
    else:
        pass # TODO
    
    
def explain_ew(fsm, state, phi, psi):
    """
    Explain why state of fsm satisfies E phi W psi.
    
    fsm -- the fsm;
    state -- a state of fsm satisfying E phi W psi;
    phi -- the set of states of fsm satifying phi;
    psi -- the set of states of fsm satifying psi.
    """
    if state <= eg(fsm, phi):
        return explain_eg(fsm, state, phi)
    else: # state <= eu(fsm, phi, psi)
        return explain_eu(fsm, state, phi, psi)
    
    
def explain_ax(fsm, state, phi):
    """
    Explain why state of fsm satisfies AX phi.
    
    fsm -- the fsm;
    state -- a state of fsm satisfying AX phi;
    phi -- the set of states of fsm satifying phi.
    """
    # To show that state satisfies AX phi, we have to show that all successors
    # satisfy phi or are not fair.
    
    # AX phi = ~fsm.pre((~phi & fair))
    
    # We don't explain why the extracted successors are not fair (when it is
    # the case).
    
    # phi or not fair state
    phi = (phi | ~fair_states(fsm)) & fsm.reachable_states
    
    # Get successor of state belonging to phi
    all_succ = explain_all_succ(fsm, state, phi)
    
    # Build successors
    successors = set()
    for inputs, next in all_succ:
        succ.add((inputs, Explanation(next)))
    
    return Explanation(state, succ)
    
    
def explain_au(fsm, state, phi, psi, states=None):
    """
    Explain why state of fsm satisfies A phi U psi.
    
    fsm -- the fsm;
    state -- a state of fsm satisfying A phi U psi;
    phi -- the set of states of fsm satifying phi;
    psi -- the set of states of fsm satifying psi;
    states -- a dictionary of state->explanation pairs.
    """
    pass # TODO
    
    
def explain_ag(fsm, state, phi):
    """
    Explain why state of fsm satisfies AG phi.
    
    fsm -- the fsm;
    state -- a state of fsm satisfying AG phi;
    phi -- the set of states of fsm satifying phi.
    """
    pass # TODO
    
    
def explain_aw(fsm, state, phi, psi):
    """
    Explain why state of fsm satisfies A phi W psi.
    
    fsm -- the fsm;
    state -- a state of fsm satisfying A phi W psi;
    phi -- the set of states of fsm satifying phi;
    psi -- the set of states of fsm satifying psi.
    """
    pass # TODO
    
    
def explain_fair(fsm, state):
    """
    Explain why state of fsm is a fair state.
    
    fsm -- the fsm;
    state -- a fair state of fsm.
    """
    return explain_eg(fsm, state, BDD.true(fsm.bddEnc.DDmanager))
    
    
def explain_not_fair(fsm, state):
    """
    Explain why state of fsm is not a fair state.
    
    fsm -- the fsm;
    state -- a non-fair state of fsm.
    """
    pass # TODO


"""
This module contains the class definition of the symbols and grammar of an LTL
formula
"""

import pyparsing as pp
import pynusmv_tools.bmcLTL.ast as ast

###############################################################################
# Symbols and keywords
###############################################################################

# terminals
number   = pp.Regex("[0-9]+").setParseAction(lambda s,l,t: ast.Constant(t[0]))
variable = pp.Regex("[a-zA-Z_@]+[a-zA-Z0-9_@.]*").setParseAction(lambda s,l,t: ast.Proposition(t[0]))
true     = pp.Keyword("TRUE" ).setParseAction(lambda s,l,t: ast.Constant("TRUE"))
false    = pp.Keyword("FALSE").setParseAction(lambda s,l,t: ast.Constant("FALSE"))

# temporal operators
op_G     = pp.Literal("[]")
op_F     = pp.Literal("<>")
op_X     = pp.Literal("()")
op_U     = pp.Keyword("U")
op_W     = pp.Keyword("W")

# propositional operators
op_not   = pp.Literal("!")
op_and   = pp.Literal("&")
op_or    = pp.Literal("|")
op_xor   = pp.Literal("^")
op_impl  = pp.Literal("=>")
op_iff   = pp.Literal("<=>")


# comparison
op_le    = pp.Literal("<=")
op_ge    = pp.Literal(">=")
op_lt    = pp.Literal("<")
op_gt    = pp.Literal(">")
op_eq    = pp.Literal("=")
op_neq   = pp.Literal("!=")

# arithmetic
op_lshift= pp.Literal("<<")
op_rshift= pp.Literal(">>")
op_add   = pp.Literal("+")
op_sub   = pp.Literal("-")
op_mul   = pp.Literal("*")
op_div   = pp.Literal("/")
op_mod   = pp.Literal("mod")

###############################################################################
# Parse actions used to instantiate meaningful 'nodes' of the ast.
# Note: the 'node' and the 'ast' are notions similar to but different from 
#       those in NuSMV.  
###############################################################################

# Just a map to decide based on the operator type the kind of node which needs
# to be instantiated.
factory = {
    "[]"   : ast.Globally,
    "<>"   : ast.Eventually,
    "()"   : ast.Next,
    "!"    : ast.Not,
    "U"    : ast.Until,
    "W"    : ast.WeakUntil,
    "&"    : ast.And,
    "|"    : ast.Or,
    "^"    : ast.Xor,
    "=>"   : ast.Imply,
    "<=>"  : ast.Equiv
}

def instanciate(s,l,tokens):
    """
    As the name suggests, this parse action is used to instantiate the ast node
    corresponding to some parsed expression (the tokens are passed as `tokens`)
    
    .. note::
        You should really not worry to much about the meaning of the parameters,
        there are inherited from the parseAction protocol.
        
    :param s: the original string corresponding to the text that has been parsed
    :param l: the location in the string where matching started
    :param tokens: the list of the matched tokens 
        (actually a :see:`pyparsing.ParseResult`)
    :return: the ast node (as defined in :see:`pynusmv_tools.bmcLTL.ast` corresponding 
        to the expression that has been parsed. 
    """
    members = tokens[0]
    if len(members)==2:
        return factory[members[0]](members[1])
    elif len(members) == 3:
        return factory[members[1]](members[0], members[2])
    else:
        # special case when the same binary operator is chained many times
        while len(members) > 1:
            accumulator = factory[members[1]](members[0], members[2])
            members     = [accumulator] + members[3:]
        return members[0]

def make_plain_proposition(s,l,tokens):
    """
    This parse action makes a plain Proposition (as defined in 
    :see:`pynusmv_tools.bmcLTL.ast.Proposition`) from the parsed text. 
    
    .. note::
        This action 'flattens' the expression; meaning that the result of the
        parsing of a complex proposition such as '(a + b) << 4 >= c - 6' will
        result in one single `Proposition` blob instead of being decomposed. 
        Namely, the previous example would yield::
        
            ast.Proposition( (a + b) << 4 >= c - 6 )
        
        This is a design choice motivated by the fact that:
        
            1. the objective of this tool is to illustrate the functioning of
               LTL bounded model checking and how its reduction to propositional 
               satisfiability (SAT) can be implemented with PyNuSMV as if it 
               were a brand new logical formalism; not to demonstrate how 
                relational and arithmetic expressions can be booleanized.
            2. the expressions represented this way do not contain any of the
               logical connectives which are part of LTL syntax and are as such
               out of the scope of this tool. (They can have no impact on the
               generated BMC problem).
            3. Relational and arithmetic 
               operators are only provided for the sake of being able to perform
               useful verification on realistic like models without being forced
               to go through the burden of defining DEFINE for each of the 
               expressions used.
               
    .. note::
        You should really not worry to much about the meaning of the parameters,
        there are inherited from the parseAction protocol.
        
    :param s: the original string corresponding to the text that has been parsed
    :param l: the location in the string where matching started
    :param tokens: the list of the matched tokens (actually a 
        :see:`pyparsing.ParseResult`)
    :return: a `Proposition` ast node encapsulating the parsed expression. 
    """
    members = tokens[0]
    text_of = lambda x: '('+x.id+')' if hasattr(x, 'id') else x
    members = list(map(text_of, members))
    return ast.Proposition(' '.join(members))

###############################################################################
# Grammar rules
###############################################################################

# The atoms (terminals) that may appear in a formula 
atom  = (true ^ false ^ variable ^ number)

# The arithmetic sub grammar: it implements the operators prececdence the usual 
# way. Concretely, we have:
#
# -- High precedence
#
# * unary minus
# * multiplication, division, modulus
# * addition, subtraction
# * left shift, right shift
#
# -- Low  precedence
#
arithmetic = pp.infixNotation(atom,
  [
   ((op_sub)                     , 1, pp.opAssoc.RIGHT,make_plain_proposition),
   ((op_mul    | op_div | op_mod), 2, pp.opAssoc.LEFT, make_plain_proposition),
   ((op_add    | op_sub),          2, pp.opAssoc.LEFT, make_plain_proposition),
   ((op_lshift | op_rshift),       2, pp.opAssoc.LEFT, make_plain_proposition),
  ])

# The relational sub grammar: it provides the <, <=, >, >=, = and != operators
comparison = pp.infixNotation(arithmetic,
  [
   ((op_le | op_lt | op_ge | op_gt | op_eq | op_neq), 2, pp.opAssoc.LEFT, make_plain_proposition)
  ])

# The rules to parse the LTL syntax proper and build a meaningful AST out of it.
# 
# Note: operator precedence was inspired by that descibed here:
#       http://vlsi.colorado.edu/~vis/doc/ctl_ltl/ctl_ltl/ctl_ltl.html
#
# Concretely, we have:
# -- High precedence
#
# * not, globally, eventually, next
# * and
# * or
# * xor
# * iff
# * implies
# * until
# * weak_until
#
# -- Low  precedence
#
LTL = pp.infixNotation(comparison,
  [
  ((op_not | op_X | op_F | op_G), 1, pp.opAssoc.RIGHT , instanciate),
  (op_and,  2, pp.opAssoc.LEFT  , instanciate),
  (op_or,   2, pp.opAssoc.LEFT  , instanciate),
  (op_xor,  2, pp.opAssoc.LEFT  , instanciate),
  (op_iff,  2, pp.opAssoc.LEFT  , instanciate),
  (op_impl, 2, pp.opAssoc.LEFT  , instanciate),
  (op_U,    2, pp.opAssoc.LEFT  , instanciate),
  (op_W,    2, pp.opAssoc.LEFT  , instanciate)
  ])
 
def parseLTL(text):
    """
    Parses an LTL string and returns the corresponding AST
     
    :param text: the text to be parsed
    :return: an LTL ast corresponding to the parsed LTL expression
    """
    return LTL.parseString(text)[0]



#! /usr/local/bin/python3
"""
This module is the entry point of a simple sat based bounded model checker for
the Linear Temporal Logic (LTL) implemented using the PyNuSMV library.
"""
import sys
import argparse

from pynusmv.init         import init_nusmv
from pynusmv.glob         import load
from pynusmv.bmc.glob     import BmcSupport
from pynusmv_tools.bmcLTL.parsing import parseLTL
from pynusmv_tools.bmcLTL.check   import check_ltl

def arguments():
    """
    Creates the arguments parser and manages to react to wrong usage.

    :returns: an object having field to store each of the command line arguments
    """
    parser = argparse.ArgumentParser(description="a PyNuSMV backed LTL sat based bmc verifier for LTL")
    parser.add_argument("-v", "--verbose", action="store_true", help="Displays the text of the analyzed model")
    parser.add_argument("-k", "--bound",   type=int, default=10, help="the maximum number of steps in a verified path")
    parser.add_argument("-s", "--spec",    type=str, help="the LTL specification to verify")
    parser.add_argument("-f", "--no-fairness",  help="disable the use of fairness constraints", action="store_true")
    parser.add_argument("-i", "--no-invariants",help="disable the invariants enforcement", action="store_true")
    parser.add_argument("-d", "--dry-run", action="store_true", help="do not perform the verification (no sat solving)")
    parser.add_argument("model", type=str, help="the name of a file containing an SMV model")

    return parser.parse_args()

def check(formula, args):
    try:
        parsed_fml          = parseLTL(formula.strip())
        status,length,trace = check_ltl(parsed_fml, args.bound, args.no_fairness, args.no_invariants, args.dry_run)
        if status != 'Ok':
            print("-- {} for length {}".format(status, length))
            print(trace)
    except Exception as e:
        print("The specification contains a syntax error")
        print(e)

def main():
    """
    The main program.
    """
    args = arguments()
    with init_nusmv():
        load(args.model)
        if args.verbose:
            with open(args.model) as f:
                print(f.read())

        with BmcSupport():
            if args.spec is not None:
                check(args.spec, args)
            else:
                print("Enter LTL properties, one per line:")
                for line in sys.stdin:
                    check(line, args)

if __name__ == "__main__":
    main()



from pyparsing import Suppress, SkipTo, Forward, ZeroOrMore, Literal, Group

from .ast import (TrueExp, FalseExp,
                  Atom, Not, And, Or, Implies, Iff, 
                  CEF, CEG, CEX, CEU, CEW, CAF, CAG, CAX, CAU, CAW)
                  
                  
def _left_(clss, tokens):
    """
    Parse tokens and return an AST, assuming left associativity.
    
    Given a list of tokens [v1, op, v2, op, ..., op, vn],
    return res, a hierarchy of instances of clss such that
    res = clss(clss(...clss(v1, v2), ..., vn).
    
    This is a helper function to parse logical operators.
    """
    if len(tokens) == 1:
        return tokens[0]
    else:
        return clss(_left_(clss, tokens[:-2]), tokens[-1])


def _right_(clss, tokens):
    """
    Parse tokens and return an AST, assuming right associativity.
    
    Given a list of tokens [v1, op, v2, op, ..., op, vn],
    return res, a hierarchy of instances of clss such that
    res = clss(v1, clss(v2, ..., vn)...).
    
    This is a helper function to parse logical operators.
    """
    if len(tokens) == 1:
        return tokens[0]
    else:
        return clss(tokens[0], _right_(clss, tokens[2:]))
        
        
def _logicals_(atomic):
    """
    Return a new parser parsing logical expressions on atomics.
    
    This parser recognizes the following grammar, with precedences
    parser := atomic | '~' parser | parser '&' parser | parser '|' parser |
              parser '->' parser | parser '<->' parser
              
    Returned AST uses .ast package's classes.
    """
    parser = Forward()

    atom = (atomic | Suppress("(") + parser + Suppress(")"))

    notstrict = "~" + atom
    notstrict.setParseAction(lambda tokens: Not(tokens[1]))
    not_ = notstrict | atom
    and_ = not_ + ZeroOrMore("&" + not_)
    and_.setParseAction(lambda tokens: _left_(And, tokens))
    or_ = and_ + ZeroOrMore("|" + and_)
    or_.setParseAction(lambda tokens: _left_(Or, tokens))
    implies = ZeroOrMore(or_ + "->") + or_
    implies.setParseAction(lambda tokens: _right_(Implies, tokens))
    iff = implies + ZeroOrMore("<->" + implies)
    iff.setParseAction(lambda tokens: _left_(Iff, tokens))

    parser <<= iff
    
    return parser
        

"""
ATL parsing tool.

_atl       := _atom | _logical | _strategic
_logical    := '~' __atl | '(' _logical ')' | __atl '&' __atl |
               __atl '|' __atl | __atl '->' __atl | __atl '<->' __atl
_strategic  := '<' _group '>' 'F' __atl | '<' _group '>' 'G' __atl |
               '<' _group '>' 'X' __atl |
               '<' _group '>' '[' __atl 'U' __atl ']' |
               '<' _group '>' '[' __atl 'W' __atl ']' |
               '[' _group ']' 'F' __atl | '[' _group ']' 'G' __atl |
               '[' _group ']' 'X' __atl |
               '[' _group ']' '[' __atl 'U' __atl ']' |
               '[' _group ']' '[' __atl 'W' __atl ']'
_agent      := _atom
_group      := _agent | _agent ',' _group

               
_atom is defined by any string surrounded by single quotes.

_logical are specified with usual precedences and associativity,
i.e.
prec : ~, &, |, ->, <->
assoc : &, |, <-> left assoc, ->, ~ right assoc


The parser returns a structure embedding the structure of the parsed
expression, represented using AST classes of .ast module.
"""

__atl = None

def parseATL(spec):
    """Parse the spec and return the list of possible ASTs."""
    global __atl
    if __atl is None:
        true = Literal("True")
        true.setParseAction(lambda tokens: TrueExp())
        false = Literal("False")
        false.setParseAction(lambda tokens: FalseExp())
        
        atom = "'" + SkipTo("'") + "'"
        atom.setParseAction(lambda tokens: Atom(tokens[1]))
        
        agent = atom
        group = Group(ZeroOrMore(agent + Suppress(",")) + agent)
        
        proposition = true | false | atom
        
        __atl = Forward()

        notproposition = "~" + proposition
        notproposition.setParseAction(lambda tokens: Not(tokens[1]))
        formula = (proposition | notproposition |
                   Suppress("(") + __atl + Suppress(")"))

        logical = Forward()
        
        
        cax = Literal("[") + group + "]" + "X" + logical
        cax.setParseAction(lambda tokens: CAX(tokens[1], tokens[4]))
        cex = Literal("<") + group + ">" + "X" + logical
        cex.setParseAction(lambda tokens: CEX(tokens[1], tokens[4]))
         
        caf = Literal("[") + group + "]" + "F" + logical
        caf.setParseAction(lambda tokens: CAF(tokens[1], tokens[4]))
        cef = Literal("<") + group + ">" + "F" + logical
        cef.setParseAction(lambda tokens: CEF(tokens[1], tokens[4]))
         
        cag = Literal("[") + group + "]" + "G" + logical
        cag.setParseAction(lambda tokens: CAG(tokens[1], tokens[4]))
        ceg = Literal("<") + group + ">" + "G" + logical           
        ceg.setParseAction(lambda tokens: CEG(tokens[1], tokens[4]))
         
        cau = Literal("[") + group + "]" + "[" + __atl + "U" + __atl + "]"
        cau.setParseAction(lambda tokens: CAU(tokens[1], tokens[4], tokens[6]))
        ceu = Literal("<") + group + ">" + "[" + __atl + "U" + __atl + "]"   
        ceu.setParseAction(lambda tokens: CEU(tokens[1], tokens[4], tokens[6]))
                                                                
        caw = Literal("[") + group + "]" + "[" + __atl + "W" + __atl + "]"   
        caw.setParseAction(lambda tokens: CAW(tokens[1], tokens[4], tokens[6]))
        cew = Literal("<") + group + ">" + "[" + __atl + "W" + __atl + "]"   
        cew.setParseAction(lambda tokens: CEW(tokens[1], tokens[4], tokens[6]))

        strategic = (cax | cex | caf | cef | cag | ceg | cau | ceu | caw | cew)
        
        logical <<= (formula | strategic)

        __atl <<= (_logicals_(logical))
    
    return __atl.parseString(spec, parseAll = True)


"""
ATLK with full observability evaluation functions.
"""

from pynusmv.dd import BDD
from pynusmv.mc import eval_simple_expression
from pynusmv.utils import fixpoint as fp

from .ast import (TrueExp, FalseExp, Init, Reachable,
                  Atom, Not, And, Or, Implies, Iff, 
                  AF, AG, AX, AU, AW, EF, EG, EX, EU, EW,
                  nK, nE, nD, nC, K, E, D, C,
                  CEF, CEG, CEX, CEU, CEW, CAF, CAG, CAX, CAU, CAW)


def evalATLK(fsm, spec):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    
    fsm -- a MAS representing the system
    spec -- an AST-based ATLK specification
    """
    
    if type(spec) is TrueExp:
        return BDD.true(fsm.bddEnc.DDmanager)
        
    elif type(spec) is FalseExp:
        return BDD.false(fsm.bddEnc.DDmanager)
        
    elif type(spec) is Init:
        return fsm.init
        
    elif type(spec) is Reachable:
        return fsm.reachable_states
    
    elif type(spec) is Atom:
        return eval_simple_expression(fsm, spec.value)
        
    elif type(spec) is Not:
        return ~evalATLK(fsm, spec.child)
        
    elif type(spec) is And:
        return evalATLK(fsm, spec.left) & evalATLK(fsm, spec.right)
        
    elif type(spec) is Or:
        return evalATLK(fsm, spec.left) | evalATLK(fsm, spec.right)
        
    elif type(spec) is Implies:
        # a -> b = ~a | b
        return (~evalATLK(fsm, spec.left)) | evalATLK(fsm, spec.right)
        
    elif type(spec) is Iff:
        # a <-> b = (a & b) | (~a & ~b)
        l = evalATLK(fsm, spec.left)
        r = evalATLK(fsm, spec.right)
        return (l & r) | ((~l) & (~r))
        
    elif type(spec) is EX:
        return ex(fsm, evalATLK(fsm, spec.child))
        
    elif type(spec) is AX:
        # AX p = ~EX ~p
        return ~ex(fsm, ~evalATLK(fsm, spec.child))
        
    elif type(spec) is EG:
        return eg(fsm, evalATLK(fsm, spec.child))
        
    elif type(spec) is AG:
        # AG p = ~EF ~p = ~E[ true U ~p ]
        return ~eu(fsm,
                   BDD.true(fsm.bddEnc.DDmanager),
                   ~evalATLK(fsm, spec.child))
        
    elif type(spec) is EU:
        return eu(fsm, evalATLK(fsm, spec.left), evalATLK(fsm, spec.right))
        
    elif type(spec) is AU:
        # A[p U q] = ~E[~q W ~p & ~q] = ~(E[~q U ~p & ~q] | EG ~q)
        p = evalATLK(fsm, spec.left)
        q = evalATLK(fsm, spec.right)
        equpq = eu(fsm, ~q, ~q & ~p)
        egq = eg(fsm, ~q)
        return ~(equpq | egq)
        
    elif type(spec) is EF:
        # EF p = E[ true U p ]
        return eu(fsm,
                  BDD.true(fsm.bddEnc.DDmanager),
                  evalATLK(fsm, spec.child))    
        
    elif type(spec) is AF:
        # AF p = ~EG ~p
        return ~eg(fsm, ~evalATLK(fsm, spec.child))
        
    elif type(spec) is EW:
        # E[ p W q ] = E[ p U q ] | EG p
        return (eu(fsm, evalATLK(fsm, spec.left), evalATLK(fsm, spec.right)) |
                eg(fsm, evalATLK(fsm, spec.left)))
        
    elif type(spec) is AW:
        # A[p W q] = ~E[~q U ~p & ~q]
        p = evalATLK(fsm, spec.left)
        q = evalATLK(fsm, spec.right)
        return ~eu(fsm, ~q, ~p & ~q)
        
    elif type(spec) is nK:
        return nk(fsm, spec.agent.value, evalATLK(fsm, spec.child))
        
    elif type(spec) is K:
        # K<'a'> p = ~nK<'a'> ~p
        return ~nk(fsm, spec.agent.value, ~evalATLK(fsm, spec.child))
        
    elif type(spec) is nE:
        return ne(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child))
        
    elif type(spec) is E:
        # E<g> p = ~nE<g> ~p
        return ~ne(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child))
        
    elif type(spec) is nD:
        return nd(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child)) 
        
    elif type(spec) is D:
        # D<g> p = ~nD<g> ~p
        return ~nd(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child))
        
    elif type(spec) is nC:
        return nc(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child))
        
    elif type(spec) is C:
        # C<g> p = ~nC<g> ~p
        return ~nc(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child))
                   
    elif type(spec) is CEX:
        # <g> X p = ~[g] X ~p
        #return ~cax(fsm, {atom.value for atom in spec.group},
        #                 ~evalATLK(fsm, spec.child))
        return cex(fsm, {atom.value for atom in spec.group},
                   evalATLK(fsm, spec.child))
        
    elif type(spec) is CAX:
        return cax(fsm, {atom.value for atom in spec.group},
                        evalATLK(fsm, spec.child))
        
    elif type(spec) is CEG:
        # <g> G p = ~[g] F ~p
        #return ~cau(fsm, {atom.value for atom in spec.group},
        #            BDD.true(fsm.bddEnc.DDmanager),
        #            ~evalATLK(fsm, spec.child))
        return cew(fsm, {atom.value for atom in spec.group},
                   evalATLK(fsm, spec.child), BDD.false(fsm.bddEnc.DDmanager))
        
    elif type(spec) is CAG:
        return cag(fsm, {atom.value for atom in spec.group},
                        evalATLK(fsm, spec.child))
        
    elif type(spec) is CEU:
        # <g> p U q = ~[g][ ~q W ~p & ~q ]
        #return ~caw(fsm, {atom.value for atom in spec.group},
        #            ~evalATLK(fsm, spec.right),
        #            ~evalATLK(fsm, spec.right) & ~evalATLK(fsm, spec.left))
        return ceu(fsm, {atom.value for atom in spec.group},
                   evalATLK(fsm, spec.left), evalATLK(fsm, spec.right))
        
    elif type(spec) is CAU:
        return cau(fsm, {atom.value for atom in spec.group},
                        evalATLK(fsm, spec.left),
                        evalATLK(fsm, spec.right))
        
    elif type(spec) is CEF:
        # <g> F p = ~[g] G ~p
        #return ~cag(fsm, {atom.value for atom in spec.group},
        #                 ~evalATLK(fsm, spec.child))
        return ceu(fsm, {atom.value for atom in spec.group},
                   BDD.true(fsm.bddEnc.DDmanager), evalATLK(fsm, spec.child))
        
    elif type(spec) is CAF:
        # [g] F p = [g][true U p]
        return cau(fsm, {atom.value for atom in spec.group},
                        BDD.true(fsm.bddEnc.DDmanager),
                        evalATLK(fsm, spec.child))
        
    elif type(spec) is CEW:
        # <g>[p W q] = ~[g][~q U ~p & ~q]
        #return ~cau(fsm, {atom.value for atom in spec.group},
        #                 ~evalATLK(fsm, spec.right),
        #                 ~evalATLK(fsm, spec.right) & ~evalATLK(fsm, spec.left))
        return cew(fsm, {atom.value for atom in spec.group},
                   evalATLK(fsm, spec.left), evalATLK(fsm, spec.right))
        
    elif type(spec) is CAW:
        return caw(fsm, {atom.value for atom in spec.group},
                        evalATLK(fsm, spec.left),
                        evalATLK(fsm, spec.right))
        
    else:
        # TODO Generate error
        print("[ERROR] evalATLK: unrecognized specification type", spec)
        return None
    
def fair_states(fsm):
    """
    Return the set of fair states of the model.
    
    fsm - the model
    """
    return eg(fsm, BDD.true(fsm.bddEnc.DDmanager))
    
    
def ex(fsm, phi):
    """
    Return the set of states of fsm satisfying EX phi.
    
    fsm -- a MAS representing the system
    phi -- a BDD representing the set of states of fsm satisfying phi
    """
    
    phi = phi.forsome(fsm.bddEnc.inputsCube) & fsm.bddEnc.statesMask
    return fsm.pre(phi & fair_states(fsm))
    
    
def eg(fsm, phi):
    """
    Return the set of states of fsm satisfying EG phi.
    
    fsm -- a MAS representing the system
    phi -- a BDD representing the set of states of fsm satisfying phi
    """
    #def inner(Z):
    #    res = Z
    #    for f in fsm.fairness_constraints:
    #        res = res & fp(lambda Y : (Z & f) | (phi & fsm.weak_pre(Y)),
    #                             BDD.false(fsm.bddEnc.DDmanager))
    #    return phi & fsm.weak_pre(res)
    #    
    #r = fp(inner, BDD.true(fsm.bddEnc.DDmanager))
    #return r.forsome(fsm.bddEnc.inputsCube)
    
    phi = phi.forsome(fsm.bddEnc.inputsCube) & fsm.bddEnc.statesMask
    
    if len(fsm.fairness_constraints) == 0:
        return fp(lambda Z : phi & fsm.pre(Z),
                  BDD.true(fsm.bddEnc.DDmanager)).forsome(fsm.bddEnc.inputsCube)
    else:
        def inner(Z):
            res = phi
            for f in fsm.fairness_constraints:
                res = res & fsm.pre(fp(lambda Y : (Z & f) |
                                       (phi & fsm.pre(Y)),
                                       BDD.false(fsm.bddEnc.DDmanager)))
            return res
        return (fp(inner, BDD.true(fsm.bddEnc.DDmanager))
                .forsome(fsm.bddEnc.inputsCube))
    
    
def eu(fsm, phi, psi):
    """
    Return the set of states of fsm satisfying E[ phi U psi ].
    
    fsm -- a MAS representing the system
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    """
    
    phi = phi.forsome(fsm.bddEnc.inputsCube) & fsm.bddEnc.statesMask
    psi = psi.forsome(fsm.bddEnc.inputsCube) & fsm.bddEnc.statesMask
    return fp(lambda X : (psi & fair_states(fsm) & fsm.reachable_states) |
                               (phi & ex(fsm, X)),
                    BDD.false(fsm.bddEnc.DDmanager))
    
    
def nk(fsm, agent, phi):
    """
    Return the set of states of fsm satisfying nK<'agent'> phi
    
    fsm -- a MAS representing the system
    agent -- the (str) name of an agent of fsm
    phi -- a BDD representing the set of states of fsm satisfying phi
    """
    # Return the set of states that have a successor in agent's knowledge
    # that satisfies phi and is reachable
    # nK<'a'> 'p' = fsm.equivalent_states(phi, agent)
    return fsm.equivalent_states(phi & fsm.reachable_states & fair_states(fsm),
                                 frozenset({agent}))
    

def ne(fsm, group, phi):
    """
    Return the set of states of fsm satisfying nE<group> phi
    
    fsm -- a MAS representing the system
    group -- a non-empty list of (str) names of agents of fsm
    phi -- a BDD representing the set of states of fsm satisfying phi
    """
    # A state satisfies nE<g> p iff there is an equivalent state in any relation
    # of any agent in g that satisfies p
    # nE<g> p = \/_{ag in g} nK<ag> p
    result = BDD.false(fsm.bddEnc.DDmanager)
    for agent in group:
        result = result | nk(fsm, agent, phi)
    return result
    
    
def nd(fsm, group, phi):
    """
    Return the set of states of fsm satisfying nD<group> phi
    
    fsm -- a MAS representing the system
    group -- a non-empty list of (str) names of agents of fsm
    phi -- a BDD representing the set of states of fsm satisfying phi
    """
    return fsm.equivalent_states(phi & fsm.reachable_states & fair_states(fsm),
                                 frozenset(group))
    
    
def nc(fsm, group, phi):
    """
    Return the set of states of fsm satisfying nC<group> phi
    
    fsm -- a MAS representing the system
    group -- a non-empty list of (str) names of agents of fsm
    phi -- a BDD representing the set of states of fsm satisfying phi
    """
    # nC<g> p = mu Z. p | nE<g> Z
    #return fp(lambda Z: (phi | ne(fsm, group, Z)),
    #          BDD.false(fsm.bddEnc.DDmanager))
    # nC<g> p = mu Z. nE<g> (p | Z)
    return fp(lambda Z: ne(fsm, group, (Z | phi)),
              BDD.false(fsm.bddEnc.DDmanager))
              
              
def cax(fsm, agents, phi):
    """
    Return the set of states of fsm satisfying [agents] X phi.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    """
    return fsm.pre_nstrat(phi & fair_gamma_states(fsm, agents), agents)
    

def cau(fsm, agents, phi, psi):
    """
    Return the set of states of fsm satisfying [agents][phi U psi].
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    """
    return fp(lambda Y : (psi & fair_gamma_states(fsm, agents)) |
                         (phi & fsm.pre_nstrat(Y, agents)),
              BDD.false(fsm.bddEnc.DDmanager))
    

def caw(fsm, agents, phi, psi):
    """
    Return the set of states of fsm satisfying [agents][phi W psi].
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    """
    if len(fsm.fairness_constraints) == 0:
        return fp(lambda Z : psi | (phi & fsm.pre_nstrat(Z, agents)),
                  BDD.true(fsm.bddEnc.DDmanager))
    else:
        def inner(Z):
            res = phi
            for f in fsm.fairness_constraints:
                res = res & fsm.pre_nstrat(fp(lambda Y :
                                             (psi &fair_gamma_states(fsm,
                                                                     agents)) |
                                             (Z & f) |
                                             (phi & fsm.pre_nstrat(Y, agents)),
                                             BDD.false(fsm.bddEnc.DDmanager)),
                                           agents)
            return (psi &fair_gamma_states(fsm, agents)) | res
        return fp(inner, BDD.true(fsm.bddEnc.DDmanager))
    
    
def cag(fsm, agents, phi):
    """
    Return the set of states of fsm satisfying [agents] G phi.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    """
    if len(fsm.fairness_constraints) == 0:
        return fp(lambda Z : phi & fsm.pre_nstrat(Z, agents),
                  BDD.true(fsm.bddEnc.DDmanager))
    else:
        def inner(Z):
            res = phi
            for f in fsm.fairness_constraints:
                res = res & fsm.pre_nstrat(fp(lambda Y : (Z & f) |
                                             (phi & fsm.pre_nstrat(Y, agents)),
                                             BDD.false(fsm.bddEnc.DDmanager))
                                           , agents)
            return res
        return fp(inner, BDD.true(fsm.bddEnc.DDmanager))
    
    
__fair_gamma_states = {}
def fair_gamma_states(fsm, agents):
    """
    Return the set of states in which agents cannot avoid a fair path.
    
    fsm -- the model
    agents -- a list of agents names
    """
    return cag(fsm, agents, BDD.true(fsm.bddEnc.DDmanager))
    
    # WARNING !!! The following code performs a segmentation fault
    # The problem seems to come from the condition
    #global _fair_gamma_states
    #agents = frozenset(agents)
    #if agents not in _fair_gamma_states:
    #    _fair_gamma_states[agents] = cag(fsm, agents,
    #                                     BDD.true(fsm.bddEnc.DDmanager))
    #return _fair_gamma_states[agents]
    
    
def cex(fsm, agents, phi):
    """
    Return the set of states of fsm satisfying <agents> X phi.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    """
    #phi = phi & fsm.bddEnc.statesInputsMask
    
    return fsm.pre_strat(phi | nfair_gamma_states(fsm, agents), agents)
    

def ceu(fsm, agents, phi, psi):
    """
    Return the set of states of fsm satisfying <agents>[phi U psi].
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    """
    #phi = phi & fsm.bddEnc.statesInputsMask
    #psi = psi & fsm.bddEnc.statesInputsMask
    
    if len(fsm.fairness_constraints) == 0:
        return fp(lambda Z : psi | (phi & fsm.pre_strat(Z, agents)),
                  BDD.false(fsm.bddEnc.DDmanager))
    else:
        nfair = nfair_gamma_states(fsm, agents)
        def inner(Z):
            res = psi
            for f in fsm.fairness_constraints:
                nf = ~f #& fsm.bddEnc.statesMask
                res = res | fsm.pre_strat(fp(lambda Y :
                                             (phi | psi | nfair) &
                                              (Z | nf) &
                                              (psi | fsm.pre_strat(Y, agents)),
                                              BDD.true(fsm.bddEnc.DDmanager)),
                                              agents)
            return (psi | phi | nfair) & res
        return fp(inner, BDD.false(fsm.bddEnc.DDmanager))
    

def cew(fsm, agents, phi, psi):
    """
    Return the set of states of fsm satisfying [agents][phi W psi].
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    """
    #phi = phi & fsm.bddEnc.statesInputsMask
    #psi = psi & fsm.bddEnc.statesInputsMask
    
    nfair = nfair_gamma_states(fsm, agents)
    
    return fp(lambda Y : (psi | phi | nfair) &
                         (psi | fsm.pre_strat(Y, agents)),
              BDD.true(fsm.bddEnc.DDmanager))
    
    
__nfair_gamma_states = {}
def nfair_gamma_states(fsm, agents):
    """
    Return the set of states in which agents cann avoid fair paths.
    
    fsm -- the model
    agents -- a list of agents names
    """
    # NFair_Gamma = not([Gamma] G True) = <Gamma> F False
    agents = frozenset(agents)
    if agents not in __nfair_gamma_states:
        if len(fsm.fairness_constraints) == 0:
            __nfair_gamma_states[agents] = BDD.false(fsm.bddEnc.DDmanager)
        else:
            def inner(Z):
                res = BDD.false(fsm.bddEnc.DDmanager)
                for f in fsm.fairness_constraints:
                    nf = ~f #& fsm.bddEnc.statesMask
                    res = res | fsm.pre_strat(fp(lambda Y :
                                                 (Z | nf) &
                                                 fsm.pre_strat(Y, agents),
                                                BDD.true(fsm.bddEnc.DDmanager)),
                                              agents)
                return res
            __nfair_gamma_states[agents] = fp(inner, 
                                              BDD.false(fsm.bddEnc.DDmanager))
    return __nfair_gamma_states[agents]



"""
ATL evaluation functions.
"""

from pynusmv.dd import BDD
from pynusmv.mc import eval_simple_expression
from pynusmv.utils import fixpoint as fp

from .ast import (TrueExp, FalseExp,
                  Atom, Not, And, Or, Implies, Iff,
                  CEF, CEG, CEX, CEU, CEW, CAF, CAG, CAX, CAU, CAW)


def evalATL(fsm, spec):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    
    fsm -- a MAS representing the system
    spec -- an AST-based ATL specification
    """
    
    if type(spec) is TrueExp:
        return BDD.true(fsm.bddEnc.DDmanager)
        
    elif type(spec) is FalseExp:
        return BDD.false(fsm.bddEnc.DDmanager)
    
    elif type(spec) is Atom:
        return eval_simple_expression(fsm, spec.value)
        
    elif type(spec) is Not:
        return ~evalATL(fsm, spec.child)
        
    elif type(spec) is And:
        return evalATL(fsm, spec.left) & evalATL(fsm, spec.right)
        
    elif type(spec) is Or:
        return evalATL(fsm, spec.left) | evalATL(fsm, spec.right)
        
    elif type(spec) is Implies:
        # a -> b = ~a | b
        return (~evalATL(fsm, spec.left)) | evalATL(fsm, spec.right)
        
    elif type(spec) is Iff:
        # a <-> b = (a & b) | (~a & ~b)
        l = evalATL(fsm, spec.left)
        r = evalATL(fsm, spec.right)
        return (l & r) | ((~l) & (~r))
                   
    elif type(spec) is CEX:
        return cex(fsm, {atom.value for atom in spec.group},
                   evalATL(fsm, spec.child))
        
    elif type(spec) is CAX:
        # [g] X p = ~<g> X ~p
        return ~cex(fsm, {atom.value for atom in spec.group},
                    ~evalATL(fsm, spec.child))
        
    elif type(spec) is CEG:
        return ceg(fsm, {atom.value for atom in spec.group},
                   evalATL(fsm, spec.child))
        
    elif type(spec) is CAG:
        # [g] G p = ~<g> F ~p
        return ~ceu(fsm, {atom.value for atom in spec.group},
                    BDD.true(fsm.bddEnc.DDmanager),
                    ~evalATL(fsm, spec.child))
        
    elif type(spec) is CEU:
        return ceu(fsm, {atom.value for atom in spec.group},
                   evalATL(fsm, spec.left),
                   evalATL(fsm, spec.right))
        
    elif type(spec) is CAU:
        # [g] p U q = ~<g>[ ~q W ~p & ~q ]
        return ~cew(fsm, {atom.value for atom in spec.group},
                    ~evalATL(fsm, spec.right),
                    ~evalATL(fsm, spec.right) & ~evalATL(fsm, spec.left))
        
    elif type(spec) is CEF:
        # <g> F p = <g>[true U p]
        return ceu(fsm, {atom.value for atom in spec.group},
                   BDD.true(fsm.bddEnc.DDmanager),
                   evalATL(fsm, spec.child))
        
    elif type(spec) is CAF:
        # [g] F p = ~<g> G ~p
        return ~ceg(fsm, {atom.value for atom in spec.group},
                    ~evalATL(fsm, spec.child))
        
    elif type(spec) is CEW:
        return cew(fsm, {atom.value for atom in spec.group},
                   evalATL(fsm, spec.left),
                   evalATL(fsm, spec.right))
        
    elif type(spec) is CAW:
        # [g][p W q] = ~<g>[~q U ~p & ~q]
        return ~ceu(fsm, {atom.value for atom in spec.group},
                    ~evalATL(fsm, spec.right),
                    ~evalATL(fsm, spec.right) & ~evalATL(fsm, spec.left))
        
    else:
        # TODO Generate error
        print("[ERROR] evalATL: unrecognized specification type", spec)
        return None
              
              
def cex(fsm, agents, phi):
    """
    Return the set of states of fsm satisfying <agents> X phi.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    """
    return fsm.pre_strat(phi, agents)
    

def ceu(fsm, agents, phi, psi):
    """
    Return the set of states of fsm satisfying <agents>[phi U psi].
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    """
    return fp(lambda Y : psi | (phi & fsm.pre_strat(Y, agents)),
              BDD.false(fsm.bddEnc.DDmanager))
    

def cew(fsm, agents, phi, psi):
    """
    Return the set of states of fsm satisfying <agents>[phi W psi].
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    """
    return fp(lambda Z : psi | (phi & fsm.pre_strat(Z, agents)),
              BDD.true(fsm.bddEnc.DDmanager))
    
    
def ceg(fsm, agents, phi):
    """
    Return the set of states of fsm satisfying <agents> G phi.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    """
    return fp(lambda Z : phi & fsm.pre_strat(Z, agents),
              BDD.true(fsm.bddEnc.DDmanager))


"""
ATLK with partial observability evaluation functions.
"""

from pynusmv.dd import BDD
from pynusmv.mc import eval_simple_expression
from pynusmv.utils import fixpoint as fp
from pynusmv.exception import PyNuSMVError

from ..atlkFO.ast import (TrueExp, FalseExp, Init, Reachable,
                          Atom, Not, And, Or, Implies, Iff, 
                          AF, AG, AX, AU, AW, EF, EG, EX, EU, EW,
                          nK, nE, nD, nC, K, E, D, C,
                          CEF, CEG, CEX, CEU, CEW, CAF, CAG, CAX, CAU, CAW)
                          

from ..atlkFO.eval import (fair_states, ex, eg, eu, nk, ne, nd, nc)


def evalATLK(fsm, spec, variant="SF", semantics="group"):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    
    fsm -- a MAS representing the system
    spec -- an AST-based ATLK specification
    variant -- the variant of the algorithm to evaluate strategic operators;
               must be
               * "SF" for the standard way: splitting in uniform strategies then
                 filtering winning states,
               * "FS" for the alternating way: filtering winning states, then
                 splitting one conflicting equivalence class, then recurse
               * "FSF" for the filter-split-filter way: filtering winning states
                 then splitting all remaining actions into uniform strategies,
                 then filtering final winning states.
                 
    If variant is not in {"SF", "FS", "FSF"}, the standard "SF" way is used.
    """
    
    if semantics != "group":
        raise PyNuSMVError("Naive evalATLK: unsupported semantics:" +
                           semantics)
    
    if type(spec) is TrueExp:
        return BDD.true(fsm.bddEnc.DDmanager)
        
    elif type(spec) is FalseExp:
        return BDD.false(fsm.bddEnc.DDmanager)
        
    elif type(spec) is Init:
        return fsm.init
        
    elif type(spec) is Reachable:
        return fsm.reachable_states
    
    elif type(spec) is Atom:
        return eval_simple_expression(fsm, spec.value)
        
    elif type(spec) is Not:
        return ~evalATLK(fsm, spec.child, variant=variant)
        
    elif type(spec) is And:
        return (evalATLK(fsm, spec.left, variant=variant)
                & evalATLK(fsm, spec.right, variant=variant))
        
    elif type(spec) is Or:
        return (evalATLK(fsm, spec.left, variant=variant)
                | evalATLK(fsm, spec.right, variant=variant))
        
    elif type(spec) is Implies:
        # a -> b = ~a | b
        return ((~evalATLK(fsm, spec.left, variant=variant))
                 | evalATLK(fsm, spec.right, variant=variant))
        
    elif type(spec) is Iff:
        # a <-> b = (a & b) | (~a & ~b)
        l = evalATLK(fsm, spec.left, variant=variant)
        r = evalATLK(fsm, spec.right, variant=variant)
        return (l & r) | ((~l) & (~r))
        
    elif type(spec) is EX:
        return ex(fsm, evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is AX:
        # AX p = ~EX ~p
        return ~ex(fsm, ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is EG:
        return eg(fsm, evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is AG:
        # AG p = ~EF ~p = ~E[ true U ~p ]
        return ~eu(fsm,
                   BDD.true(fsm.bddEnc.DDmanager),
                   ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is EU:
        return eu(fsm, evalATLK(fsm, spec.left, variant=variant),
                       evalATLK(fsm, spec.right, variant=variant))
        
    elif type(spec) is AU:
        # A[p U q] = ~E[~q W ~p & ~q] = ~(E[~q U ~p & ~q] | EG ~q)
        p = evalATLK(fsm, spec.left, variant=variant)
        q = evalATLK(fsm, spec.right, variant=variant)
        equpq = eu(fsm, ~q, ~q & ~p)
        egq = eg(fsm, ~q)
        return ~(equpq | egq)
        
    elif type(spec) is EF:
        # EF p = E[ true U p ]
        return eu(fsm,
                  BDD.true(fsm.bddEnc.DDmanager),
                  evalATLK(fsm, spec.child, variant=variant))    
        
    elif type(spec) is AF:
        # AF p = ~EG ~p
        return ~eg(fsm, ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is EW:
        # E[ p W q ] = E[ p U q ] | EG p
        return (eu(fsm, evalATLK(fsm, spec.left, variant=variant),
                        evalATLK(fsm, spec.right, variant=variant)) |
                eg(fsm, evalATLK(fsm, spec.left, variant=variant)))
        
    elif type(spec) is AW:
        # A[p W q] = ~E[~q U ~p & ~q]
        p = evalATLK(fsm, spec.left, variant=variant)
        q = evalATLK(fsm, spec.right, variant=variant)
        return ~eu(fsm, ~q, ~p & ~q)
        
    elif type(spec) is nK:
        return nk(fsm, spec.agent.value, evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is K:
        # K<'a'> p = ~nK<'a'> ~p
        return ~nk(fsm, spec.agent.value, ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is nE:
        return ne(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is E:
        # E<g> p = ~nE<g> ~p
        return ~ne(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is nD:
        return nd(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child, variant=variant)) 
        
    elif type(spec) is D:
        # D<g> p = ~nD<g> ~p
        return ~nd(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is nC:
        return nc(fsm,
                  [a.value for a in spec.group],
                  evalATLK(fsm, spec.child, variant=variant))
        
    elif type(spec) is C:
        # C<g> p = ~nC<g> ~p
        return ~nc(fsm,
                   [a.value for a in spec.group],
                   ~evalATLK(fsm, spec.child, variant=variant))
                   
    elif type(spec) in {CEX, CAX, CEG, CAG, CEU, CAU, CEF, CAF, CEW, CAW}:
        if variant == "SF":
            return eval_strat(fsm, spec)
        elif variant == "FS":
            return eval_strat_improved(fsm, spec)
        elif variant == "FSF":
            return eval_strat_FSF(fsm, spec)
        else:
            return eval_strat(fsm, spec)
        
    else:
        # TODO Generate error
        print("[ERROR] evalATLK: unrecognized specification type", spec)
        return None



def cex(fsm, agents, phi, strat=None):
    """
    Return the set of states of strat satisfying <agents> X phi
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    strat -- a BDD representing allowed state/inputs pairs, or None
    """        
    nfair = nfair_gamma(fsm, agents, strat)
    
    return fsm.pre_strat(phi | nfair, agents, strat)
    

def ceu(fsm, agents, phi, psi, strat=None):
    """
    Return the set of states of strat satisfying <agents>[phi U psi]
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    
    if len(fsm.fairness_constraints) == 0:
        return fp(lambda Z : psi | (phi & fsm.pre_strat(Z, agents, strat)),
                  BDD.false(fsm.bddEnc.DDmanager))
    else:
        nfair = nfair_gamma(fsm, agents, strat)
        def inner(Z):
            res = psi
            for f in fsm.fairness_constraints:
                nf = ~f
                res = res | fsm.pre_strat(fp(lambda Y :
                                             (phi | psi | nfair) &
                                             (Z | nf) &
                                             (psi |
                                              fsm.pre_strat(Y, agents, strat)),
                                             BDD.true(fsm.bddEnc.DDmanager)),
                                              agents, strat)
            return (psi | phi | nfair) & res
        return fp(inner, BDD.false(fsm.bddEnc.DDmanager))
    

def cew(fsm, agents, phi, psi, strat=None):
    """
    Return the set of states of strat satisfying <agents>[phi W psi]
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """    
    nfair = nfair_gamma(fsm, agents, strat)
    
    return fp(lambda Y : (psi | phi | nfair) &
                         (psi | fsm.pre_strat(Y, agents, strat)),
              BDD.true(fsm.bddEnc.DDmanager))
    
    
def ceg(fsm, agents, phi, strat=None):
    """
    Return the set of states of strat satisfying <agents> G phi
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """    
    nfair = nfair_gamma(fsm, agents, strat)
    
    return fp(lambda Y : (phi | nfair) & fsm.pre_strat(Y, agents, strat),
              BDD.true(fsm.bddEnc.DDmanager))


def nfair_gamma(fsm, agents, strat=None):
    """
    Return the set of states of strat
    in which agents can avoid a fair path in strat.
    If strat is None, it is considered true.
    
    fsm -- the model
    agents -- a list of agents names
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """    
    if len(fsm.fairness_constraints) == 0:
        return BDD.false(fsm.bddEnc.DDmanager)
    else:
        def inner(Z):
            res = BDD.false(fsm.bddEnc.DDmanager)
            for f in fsm.fairness_constraints:
                nf = ~f
                res = res | fsm.pre_strat(fp(lambda Y :
                                              (Z | nf) &
                                              fsm.pre_strat(Y, agents, strat),
                                              BDD.true(fsm.bddEnc.DDmanager)),
                                             agents, strat)
            return res
        return fp(inner, BDD.false(fsm.bddEnc.DDmanager))
        
def split(fsm, strats, gamma):
    """
    Split strats into all its non-conflicting greatest subsets.
    
    fsm -- the model
    strats -- a BDD representing a set of states/inputs pairs
    gamma -- a set of agents of fsm
    
    """
    if strats.is_false():
        return {strats}
    
    else:
        # Get one equivalence class
        si = fsm.pick_one_state_inputs(strats)
        s = si.forsome(fsm.bddEnc.inputsCube)
        eqs = fsm.equivalent_states(s, gamma)
        eqcl = strats & eqs
    
        # Remove eqcl from strats
        strats = strats - eqcl
    
        # Get ngamma cube
        ngamma_cube = fsm.bddEnc.inputsCube - fsm.inputs_cube_for_agents(gamma)
    
        # Split eqcl into non-conflicting subsets
        # (if eqcl is already non-conflicting, only one iteration is done)
        eqcls = set()
        while eqcl.isnot_false():
            si = fsm.pick_one_state_inputs(eqcl)
            ncss = eqcl & si.forsome(fsm.bddEnc.statesCube).forsome(ngamma_cube)
            eqcls.add(ncss)
            eqcl = eqcl - ncss
        
        # Combine NCSSs with strategies of restricted strats
        unistrats = set()
        for strat in split(fsm, strats, gamma):
            for ncss in eqcls:
                unistrats.add(strat | ncss)
            
        return unistrats
    

def eval_strat(fsm, spec):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    spec is a strategic operator <G> pi.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator.
    
    """
    sat = BDD.false(fsm.bddEnc.DDmanager)
    agents = {atom.value for atom in spec.group}
    # TODO restrict protocol to reachable states to avoid splitting useless
    # equivalence classes
    strats = split(fsm, fsm.protocol(agents), agents)
    nbstrats = 0
    for strat in strats:
        nbstrats += 1
        if type(spec) is CEX:
#            winning = cex(fsm, agents, evalATLK(fsm, spec.child), strat)
            winning = cex_si(fsm, agents,
                             evalATLK(fsm, spec.child, "SF"), strat).forsome(
                      fsm.bddEnc.inputsCube)

        elif type(spec) is CAX:
            # [g] X p = ~<g> X ~p
#            winning = ~cex(fsm, agents, ~evalATLK(fsm, spec.child), strat)
            winning = ~(cex_si(fsm, agents,
                               ~evalATLK(fsm, spec.child, "SF"),
                               strat)).forsome(
                        fsm.bddEnc.inputsCube)

        elif type(spec) is CEG:
#            winning = ceg(fsm, agents, evalATLK(fsm, spec.child), strat)
            winning = ceg_si(fsm, agents,
                             evalATLK(fsm, spec.child, "SF"), strat).forsome(
                      fsm.bddEnc.inputsCube)

        elif type(spec) is CAG:
            # [g] G p = ~<g> F ~p
#            winning= ~ceu(fsm, agents, BDD.true(fsm.bddEnc.DDmanager),
#                          ~evalATLK(fsm, spec.child), strat)
            winning= ~(ceu_si(fsm, agents,
                              BDD.true(fsm.bddEnc.DDmanager),
                              ~evalATLK(fsm, spec.child, "SF"),
                              strat).forsome(
                       fsm.bddEnc.inputsCube))

        elif type(spec) is CEU:
#            winning = ceu(fsm, agents, evalATLK(fsm, spec.left),
#                          evalATLK(fsm, spec.right), strat)
            winning = ceu_si(fsm, agents,
                             evalATLK(fsm, spec.left, "SF"),
                             evalATLK(fsm, spec.right, "SF"), strat).forsome(
                      fsm.bddEnc.inputsCube)

        elif type(spec) is CAU:
            # [g][p U q] = ~<g>[ ~q W ~p & ~q ]
#            winning =  ~cew(fsm, agents, ~evalATLK(fsm, spec.right),
#                            ~evalATLK(fsm, spec.right) &
#                            ~evalATLK(fsm, spec.left), strat)
            winning =  ~(cew_si(fsm, agents,
                                ~evalATLK(fsm, spec.right, "SF"),
                                ~evalATLK(fsm, spec.right, "SF") &
                                ~evalATLK(fsm, spec.left, "SF"),
                                strat).forsome(
                         fsm.bddEnc.inputsCube))

        elif type(spec) is CEF:
            # <g> F p = <g>[true U p]
#            winning = ceu(fsm, agents, BDD.true(fsm.bddEnc.DDmanager),
#                          evalATLK(fsm, spec.child), strat)
            winning = ceu_si(fsm, agents,
                             BDD.true(fsm.bddEnc.DDmanager),
                             evalATLK(fsm, spec.child, "SF"), strat).forsome(
                      fsm.bddEnc.inputsCube)

        elif type(spec) is CAF:
            # [g] F p = ~<g> G ~p
#            winning = ~ceg(fsm, agents, ~evalATLK(fsm, spec.child), strat)
            winning = ~(ceg_si(fsm, agents,
                               ~evalATLK(fsm, spec.child, "SF"),
                               strat).forsome(
                        fsm.bddEnc.inputsCube))

        elif type(spec) is CEW:
#            winning = cew(fsm, agents, evalATLK(fsm, spec.left),
#                          evalATLK(fsm, spec.right), strat)
           winning = cew_si(fsm, agents,
                            evalATLK(fsm, spec.left, "SF"),
                            evalATLK(fsm, spec.right, "SF"), strat).forsome(
                     fsm.bddEnc.inputsCube)

        elif type(spec) is CAW:
            # [g][p W q] = ~<g>[~q U ~p & ~q]
#            winning = ~ceu(fsm, agents, ~evalATLK(fsm, spec.right),
#                           ~evalATLK(fsm, spec.right) &
#                           ~evalATLK(fsm, spec.left), strat)
            winning = ~(ceu_si(fsm, agents,
                               ~evalATLK(fsm, spec.right, "SF"),
                               ~evalATLK(fsm, spec.right, "SF") &
                               ~evalATLK(fsm, spec.left, "SF"),
                               strat).forsome(
                        fsm.bddEnc.inputsCube))
                        
        # Complete sat with states for which all states belong to winning
        
        # wineq is the set of states for which all equiv states are in winning
        nwinning = ~winning & fsm.bddEnc.statesInputsMask
        wineq = ~(fsm.equivalent_states(nwinning &
                  fsm.reachable_states, frozenset(agents))) & winning
        sat = sat | wineq
    
    # DEBUG Print number of strategies
    #print("Eval_strat: {} strategies".format(nbstrats))
    
    return sat



def cex_si(fsm, agents, phi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents> X phi
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    strat -- a BDD representing allowed state/inputs pairs, or None
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask
    
    return fsm.pre_strat_si(phi | nfair_gamma_si(fsm, agents, strat),
                            agents, strat)
    

def ceu_si(fsm, agents, phi, psi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents>[phi U psi]
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    psi = psi & fsm.bddEnc.statesInputsMask & strat
    
    if len(fsm.fairness_constraints) == 0:
        return fp(lambda Z : psi | (phi & fsm.pre_strat_si(Z, agents, strat)),
                  BDD.false(fsm.bddEnc.DDmanager))
    else:
        nfair = nfair_gamma_si(fsm, agents, strat)
        def inner(Z):
            res = psi
            for f in fsm.fairness_constraints:
                nf = ~f & fsm.bddEnc.statesMask & strat
                res = res | fsm.pre_strat_si(fp(lambda Y :
                                                 (phi | psi | nfair) &
                                                 (Z | nf) &
                                                 (psi |
                                                  fsm.pre_strat_si(Y, agents,
                                                                   strat)
                                                 ),
                                              BDD.true(fsm.bddEnc.DDmanager)),
                                              agents, strat)
            return (psi | phi | nfair) & res
        return fp(inner, BDD.false(fsm.bddEnc.DDmanager))
    

def cew_si(fsm, agents, phi, psi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents>[phi W psi]
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    psi -- a BDD representing the set of states of fsm satisfying psi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    psi = psi & fsm.bddEnc.statesInputsMask & strat
    
    nfair = nfair_gamma_si(fsm, agents, strat)
    
    return fp(lambda Y : (psi | phi | nfair) &
                         (psi | fsm.pre_strat_si(Y, agents, strat)),
              BDD.true(fsm.bddEnc.DDmanager))
    
    
def ceg_si(fsm, agents, phi, strat=None):
    """
    Return the set of state/inputs pairs of strat satisfying <agents> G phi
    under full observability in strat.
    If strat is None, strat is considered true.
    
    fsm -- a MAS representing the system
    agents -- a list of agents names
    phi -- a BDD representing the set of states of fsm satisfying phi
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    phi = phi & fsm.bddEnc.statesInputsMask & strat
    
    nfair = nfair_gamma_si(fsm, agents, strat)
    
    return fp(lambda Y : (phi | nfair) & fsm.pre_strat_si(Y, agents, strat),
              BDD.true(fsm.bddEnc.DDmanager))


def nfair_gamma_si(fsm, agents, strat=None):
    """
    Return the set of state/inputs pairs of strat
    in which agents can avoid a fair path in strat.
    If strat is None, it is considered true.
    
    fsm -- the model
    agents -- a list of agents names
    strat -- a BDD representing allowed state/inputs pairs, or None
    
    """
    if not strat:
        strat = BDD.true(fsm.bddEnc.DDmanager)
    
    if len(fsm.fairness_constraints) == 0:
        return BDD.false(fsm.bddEnc.DDmanager)
    else:
        def inner(Z):
            res = BDD.false(fsm.bddEnc.DDmanager)
            for f in fsm.fairness_constraints:
                nf = ~f & fsm.bddEnc.statesMask & strat
                res = res | fsm.pre_strat_si(fp(lambda Y :
                                                 (Z | nf) &
                                                 fsm.pre_strat_si(Y, agents,
                                                                  strat),
                                             BDD.true(fsm.bddEnc.DDmanager)),
                                             agents, strat)
            return res
        return fp(inner, BDD.false(fsm.bddEnc.DDmanager))



def eval_strat_improved(fsm, spec, strat=None):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    spec is a strategic operator <G> pi.
    
    Implement the improved algorithm alternating between splitting a conflicting
    equvalence class and filtering winning state/inputs pairs.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator.
    strat -- the strategy to consider (not necessarily uniform).
    
    """
    sat = BDD.false(fsm.bddEnc.DDmanager)
    agents = {atom.value for atom in spec.group}
    
    if not strat:
        strat = fsm.protocol(agents)
    
    if type(spec) is CEX:
        winning = cex_si(fsm, agents,
                         evalATLK(fsm, spec.child, "FS"), strat)

    elif type(spec) is CAX:
        # [g] X p = ~<g> X ~p
        newspec = CEX(spec.group, Not(spec.child))
        return ~evalATLK(fsm, newspec, "FS")

    elif type(spec) is CEG:
        winning = ceg_si(fsm, agents, evalATLK(fsm, spec.child, "FS"), strat)

    elif type(spec) is CAG:
        # [g] G p = ~<g> F ~p
        newspec = CEF(spec.group, Not(spec.child))
        return ~evalATLK(fsm, newspec, "FS")

    elif type(spec) is CEU:
        winning = ceu_si(fsm, agents, evalATLK(fsm, spec.left, "FS"),
                         evalATLK(fsm, spec.right, "FS"), strat)

    elif type(spec) is CAU:
        # [g][p U q] = ~<g>[ ~q W ~p & ~q ]
        newspec = CEW(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right)))
        return ~evalATLK(fsm, newspec, "FS")

    elif type(spec) is CEF:
        # <g> F p = <g>[true U p]
        winning = ceu_si(fsm, agents, BDD.true(fsm.bddEnc.DDmanager),
                         evalATLK(fsm, spec.child, "FS"), strat)

    elif type(spec) is CAF:
        # [g] F p = ~<g> G ~p
        newspec = CEG(spec.group, Not(spec.child))
        return ~evalATLK(fsm, newspec, "FS")

    elif type(spec) is CEW:
       winning = cew_si(fsm, agents, evalATLK(fsm, spec.left, "FS"),
                        evalATLK(fsm, spec.right, "FS"), strat)

    elif type(spec) is CAW:
        # [g][p W q] = ~<g>[~q U ~p & ~q]
        newspec = CEU(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right)))
        return ~evalATLK(fsm, newspec, "FS")
                          
    
    winning = winning & fsm.bddEnc.statesInputsMask & fsm.protocol(agents)
    # Get one conflicting equivalence class
    if winning.is_false(): # no state/inputs pairs are winning => return false
        return winning
    
    # Get ngamma cube
    gamma = agents
    ngamma_cube = fsm.bddEnc.inputsCube - fsm.inputs_cube_for_agents(gamma)
    
    conflicting = False
    orig_winning = winning
    while (not conflicting) and winning.isnot_false():
        # Get one equivalence class
        si = fsm.pick_one_state_inputs(winning)
        s = si.forsome(fsm.bddEnc.inputsCube)
        eqs = fsm.equivalent_states(s, gamma)
        eqcl = winning & eqs
        
        if ((eqcl -
             (eqcl & 
              si.forsome(fsm.bddEnc.statesCube).forsome(ngamma_cube)))
            .isnot_false()):
            # Equivalence class is conflicting because there are some other
            # possible actions, conflicting with si.
            conflicting = True
            
        else:
            # The equivalence class is not conflicting, remove it for the search
            winning = winning - eqcl
            
    # TODO Improve this: if a pair is not conflicting now, it won't be 
    # conflicting anymore. We should keep track of non-conflicting equivalence
    # classes through recursive calls
            
    if winning.is_false():
        # No conflicting classes, return states that are winning for all eq
        
        orig_winning = orig_winning.forsome(fsm.bddEnc.inputsCube)

        # wineq is the set of states for which all equiv states are in winning
        nwinning = ~orig_winning & fsm.bddEnc.statesInputsMask
        wineq = ~(fsm.equivalent_states(nwinning &
                  fsm.reachable_states, frozenset(agents))) & orig_winning
        return wineq
        
    else:
        # si is one conflicting action, eqcl is the conflicting
        # equivalence class
        # Split eqcl into non-conflicting subsets and recursively call eval
        orig_winning = orig_winning - eqcl
        eqcls = set()
        while eqcl.isnot_false():
            si = fsm.pick_one_state_inputs(eqcl)
            ncss = eqcl & si.forsome(fsm.bddEnc.statesCube).forsome(ngamma_cube)
            eqcls.add(ncss)
            eqcl = eqcl - ncss
        
        for ncss in eqcls:
            strat = orig_winning | ncss
            sat = sat | eval_strat_improved(fsm, spec, strat)
    
        return sat


def filter_strat(fsm, spec, strat=None, variant="SF"):
    """
    Returns the subset SA of strat (or the whole system if strat is None),
    state/action pairs of fsm, such that there is a strategy to satisfy spec
    in fsm.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator;
            the operator is CEX, CEG, CEF, CEU or CEW.
    strat -- the subset of the system to consider.
    variant -- the variant of the algorithm to evaluate strategic operators;
               must be
               * "SF" for the standard way: splitting in uniform strategies then
                 filtering winning states,
               * "FS" for the alternating way: filtering winning states, then
                 splitting one conflicting equivalence class, then recurse
               * "FSF" for the filter-split-filter way: filtering winning states
                 then splitting all remaining actions into uniform strategies,
                 then filtering final winning states.
                 
    If variant is not in {"SF", "FS", "FSF"}, the standard "SF" way is used.
    """
    
    sat = BDD.false(fsm.bddEnc.DDmanager)
    agents = {atom.value for atom in spec.group}
    
    # Filtering
    if type(spec) is CEX:
        winning = cex_si(fsm, agents,
                         evalATLK(fsm, spec.child, variant=variant), strat)

    elif type(spec) is CEG:
        winning = ceg_si(fsm, agents, evalATLK(fsm, spec.child, variant=variant), strat)

    elif type(spec) is CEU:
        winning = ceu_si(fsm, agents, evalATLK(fsm, spec.left, variant=variant),
                         evalATLK(fsm, spec.right, variant=variant), strat)

    elif type(spec) is CEF:
        # <g> F p = <g>[true U p]
        winning = ceu_si(fsm, agents, BDD.true(fsm.bddEnc.DDmanager),
                         evalATLK(fsm, spec.child, variant=variant), strat)

    elif type(spec) is CEW:
       winning = cew_si(fsm, agents, evalATLK(fsm, spec.left, variant=variant),
                        evalATLK(fsm, spec.right, variant=variant), strat)
    
    
    return winning & fsm.bddEnc.statesInputsMask & fsm.protocol(agents)
    
    
def eval_strat_FSF(fsm, spec):
    """
    Return the BDD representing the set of states of fsm satisfying spec.
    spec is a strategic operator <G> pi.
    
    Implement a variant of the algorithm that
    filters, splits and then filters.
    
    fsm -- a MAS representing the system;
    spec -- an AST-based ATLK specification with a top strategic operator.
    
    """
    
    if type(spec) is CAX:
        # [g] X p = ~<g> X ~p
        newspec = CEX(spec.group, Not(spec.child))
        return ~eval_strat_FSF(fsm, newspec)
        
    elif type(spec) is CAG:
        # [g] G p = ~<g> F ~p
        newspec = CEF(spec.group, Not(spec.child))
        return ~eval_strat_FSF(fsm, newspec)
        
    elif type(spec) is CAU:
        # [g][p U q] = ~<g>[ ~q W ~p & ~q ]
        newspec = CEW(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right)))
        return ~eval_strat_FSF(fsm, newspec)
        
    elif type(spec) is CAF:
        # [g] F p = ~<g> G ~p
        newspec = CEG(spec.group, Not(spec.child))
        return ~eval_strat_FSF(fsm, newspec)
        
    elif type(spec) is CAW:
        # [g][p W q] = ~<g>[~q U ~p & ~q]
        newspec = CEU(spec.group,
                      Not(spec.right),
                      And(Not(spec.left), Not(spec.right)))
        return ~eval_strat_FSF(fsm, newspec)
        
        
    sat = BDD.false(fsm.bddEnc.DDmanager)
    agents = {atom.value for atom in spec.group}
    
    # First filtering
    winning = filter_strat(fsm, spec, variant="FSF")
    
    if winning.is_false(): # no state/inputs pairs are winning => return false
        return winning
    
    # Splitting the strategies
    strats = split(fsm, winning, agents)
    nbstrats = 0
    for strat in strats:
        nbstrats += 1
        # Second filtering
        winning = filter_strat(fsm, spec, strat, variant="FSF")
        
        # wineq is the set of states for which all equiv states are in winning
        winning = winning.forsome(fsm.bddEnc.inputsCube)
        nwinning = ~winning & fsm.bddEnc.statesInputsMask
        wineq = ~(fsm.equivalent_states(nwinning &
                  fsm.reachable_states, frozenset(agents))) & winning
        sat = sat | wineq
    
    # DEBUG Print number of strategies
    #print("Eval_strat_FSF: {} strategies".format(nbstrats))
    
    return sat


from pynusmv.init import *
from pynusmv.glob import *
import sys
import argparse

"""
Compare two NuSMV model and print differences.
"""

def states_transitions(fsm):
    """
    Given a BddFsm, return the set of states and transitions of the reachable
    state-space of the fsm.
    
    The returned transitions are triplets s, i, s' telling that s leads to
    s' through i.
    If the model has no input variables, i is None.
    """
    states = fsm.pick_all_states(fsm.reachable_states)
    transitions = {(frozenset(s1.get_str_values().items()),
                    frozenset(i.get_str_values().items())
                    if i is not None else frozenset({None}),
                    frozenset(s2.get_str_values().items())
                   )
                   for s1 in states
                   for s2 in states
                   for i in (fsm.pick_all_inputs(
                                         fsm.get_inputs_between_states(s1,s2))
                             if len(fsm.bddEnc.inputsVars) > 0
                             else {None})}
    states = {frozenset(s.get_str_values().items()) for s in states}
    return states, transitions

def compare(model1, model2, comparisons=None):
    """
    Load and retrieve reachable state-space of model1 and model2 and print
    differences between the two.
    
    The compared aspects are the following:
        - states: which states are reachable in one model but not in the other.
        - common: which transitions are in one model but not in the other,
                  considering only states that are in both models.
        - transitions: which transitions (with input variables) are possible
                       in one model but not in the other.
    If comparisons is None, all statistics are printed; otherwise, only the
    ones present in comparisons are printed.
    
    model1 and model2 must be loadable by PyNuSMV.
    """
    
    if comparisons is None:
        comparisons = {"states", "common", "transitions"}
    
    init_nusmv()
    load(model1)
    compute_model()
    fsm = prop_database().master.bddFsm
    s1, t1 = states_transitions(fsm)
    reset_nusmv()
    load(model2)
    compute_model()
    fsm = prop_database().master.bddFsm
    s2, t2 = states_transitions(fsm)
    deinit_nusmv()
    
    ct12 = {(s, i, sp) for s, i, sp in t1 if s in s2}
    ct21 = {(s, i, sp) for s, i, sp in t2 if s in s1}
    
    if "states" in comparisons:
        print("=" * 79)
        print("FIRST  HAS {} REACHABLE STATES".format(len(s1)))
        print("SECOND HAS {} REACHABLE STATES".format(len(s2)))
        
        if len(s1 - s2) > 0:
            states = ["\n".join(str(var) + " = " + str(val)
                                for var, val in sorted(s))
                      for s in s1 - s2]
            
            title = ("STATES IN FIRST BUT NOT IN SECOND ({})".
                     format(len(states)))
            print("=" * 79)
            print(title)
            print("=" * 79)
            print(("\n" + "-" * 79 + "\n").join(states))
        
        if len(s2 - s1) > 0:
            states = ["\n".join(str(var) + " = " + str(val)
                                for var, val in sorted(s))
                      for s in s2 - s1]
            
            title = ("STATES IN SECOND BUT NOT IN FIRST ({})".
                     format(len(states)))
            print("=" * 79)
            print(title)
            print("=" * 79)
            print(("\n" + "-" * 79 + "\n").join(states))
    
    if "common" in comparisons:
        print("=" * 79)
        print("FIRST  HAS {} REACHABLE TRANSITIONS FROM COMMON STATES".format(
              len(ct12)))
        print("SECOND HAS {} REACHABLE TRANSITIONS FROM COMMON STATES".format(
              len(ct21)))
        
        if len(ct12 - ct21) > 0:
            transitions = ["\n".join(str(var) + " = " + str(val)
                                     for var, val in sorted(s)) +
                           (("\n".join(str(var) + " = " + str(val)
                                     for var, val in sorted(i)))
                            if None not in i else "") +
                           "\n".join(str(var) + " = " + str(val)
                                     for var, val in sorted(sp))
                      for s, i, sp in ct12 - ct21]
            
            title = ("TRANSITIONS IN FIRST BUT NOT IN SECOND ({})".
                     format(len(transitions)))
            print("=" * 79)
            print(title)
            print("=" * 79)
            print(("\n" + "-" * 79 + "\n").join(transitions))
        
        if len(ct21 - ct12) > 0:
            transitions = ["\n".join(str(var) + " = " + str(val)
                                     for var, val in sorted(s)) +
                           (("\n" + "-" * 79 + "\n" +
                            ("\n".join(str(var) + " = " + str(val)
                                     for var, val in sorted(i))))
                            if None not in i else "") +
                           "\n" + "-" * 79 + "\n" + 
                           "\n".join(str(var) + " = " + str(val)
                                     for var, val in sorted(sp))
                      for s, i, sp in ct21 - ct12]
            
            title = ("TRANSITIONS IN SECOND BUT NOT IN FIRST ({})".
                     format(len(transitions)))
            print("=" * 79)
            print(title)
            print("=" * 79)
            print(("\n" + ("-" * 79 + "\n") * 2).join(transitions))
    
    if "transitions" in comparisons:
        print("=" * 79)
        print("FIRST  HAS {} REACHABLE TRANSITIONS".format(len(t1)))
        print("SECOND HAS {} REACHABLE TRANSITIONS".format(len(t2)))
        
        if len(t1 - t2) > 0:
            transitions = ["\n".join(str(var) + " = " + str(val)
                                     for var, val in sorted(s)) +
                           (("\n".join(str(var) + " = " + str(val)
                                     for var, val in sorted(i)))
                            if None not in i else "") +
                           "\n".join(str(var) + " = " + str(val)
                                     for var, val in sorted(sp))
                      for s, i, sp in t1 - t2]
            
            title = ("TRANSITIONS IN FIRST BUT NOT IN SECOND".
                     format(len(transitions)))
            print("=" * 79)
            print(title)
            print("=" * 79)
            print(("\n" + "-" * 79 + "\n").join(transitions))
        
        if len(t2 - t1) > 0:
            transitions = ["\n".join(str(var) + " = " + str(val)
                                     for var, val in sorted(s)) +
                           (("\n" + "-" * 79 + "\n" +
                            ("\n".join(str(var) + " = " + str(val)
                                     for var, val in sorted(i))))
                            if None not in i else "") +
                           "\n" + "-" * 79 + "\n" + 
                           "\n".join(str(var) + " = " + str(val)
                                     for var, val in sorted(sp))
                      for s, i, sp in t2 - t1]
            
            title = ("TRANSITIONS IN SECOND BUT NOT IN FIRST ({})".
                     format(len(transitions)))
            print("=" * 79)
            print(title)
            print("=" * 79)
            print(("\n" + ("-" * 79 + "\n") * 2).join(transitions))

def main():
    # Parse arguments
    parser = argparse.ArgumentParser(description='SMV model comparator.')
    # Populate arguments:
    # first model, second model
    # -s show states
    # -t show transitions
    # -c show transitions from common states
    parser.add_argument('first', help='the first SMV model')
    parser.add_argument('second', help='the second SMV model')
    parser.add_argument('-s', action='store_true',
                        help="Show states comparison")
    parser.add_argument('-t', action='store_true',
                        help="Show transitions comparison")
    parser.add_argument('-c', action='store_true',
                        help="Show transitions from common states comparison")
    args = parser.parse_args(sys.argv[1:])
    
    comparisons = set()
    if args.s:
        comparisons.add("states")
    if args.t:
        comparisons.add("transitions")
    if args.c:
        comparisons.add("common")
    if len(comparisons) <= 0:
        comparisons = None


    compare(args.first, args.second, comparisons)

if __name__ == "__main__":
    main()
